(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTSymbols`"];


BeginPackage["FMTAnalytics`"];


controlFlow::usage="controlFlow[expression,highlightedSet] Returns the graph (a list of directed edges) from start to finish with branches for conditional and loop structures.\n"<>
"The highlightedSet list can be used to optionally highlight the selected points of the diagram.";


dataFlow::usage="dataFlow[expression,paramModes] Returns the list of data-flow offending operations.";


outgoingCalls::usage="outgoingCalls[expression] Returns the list of qualified operation names that are called in the given expression.";


indirectObjectUses::usage="indirectObjectUses[directUses,operationCalls] Returns closure of directUses based on the operation call graph.";


dataAliasing::usage="dataAliasing[qOperationName,expression,globalsUsed,operationsCalled] Returns the list of object aliasing detected in the expression,\n"<>
"with regard to duplicated parameters as well as passing a global object that is used by the called operation.";


(* these definitions are placed here to force all used symbols in the public context of this package: *)

analyticsTags={
	calls,
	reads,
	writes
};


Begin["`Private`"];


(* Helper, presents the object name in its shortened form,
	without array indexing, but with structure field names, if any are present *)
shortObjectName[obj_]:=
	Switch[Head[obj],
		Symbol,
			(* this is full object name, print it as is *)
			ToString[obj],
		
		FMTSymbols`objectParentT,
			(* this is composite object, either array or structure name *)
			shortObjectName[obj[[1]]],
		
		FMTSymbols`partT,
			(* this is array indexing, print as "arrayName[...]" *)
			With[
				{
					arrayObj=obj[[1]]
				},
				
				shortObjectName[arrayObj]<>"[[...]]"
			],
		
		FMTSymbols`dotT,
			(* this is structure field access, print as "structureName.fieldName" *)
			With[
				{
					structureObj=obj[[1]],
					fieldName=obj[[2]]
				},
				
				shortObjectName[structureObj]<>"."<>shortObjectName[fieldName]
			]
	]


(* Helper, reduces the fully resolved grammar to simplified structure of control flow paths *)
controlFlowReduceStep[expression_]:=Module[
	{
		replacements
	},
	
	replacements={
		(* compound constructs are just lists of nested structures: *)
		FMTSymbols`compoundT->List,
		
		(* as a special case, empty statements can be simply removed: *)
		FMTSymbols`statementT[]->Nothing,
	
		(* conditionals: *)
		
		(* ifT without the else branch: *)
		
		FMTSymbols`numberedT[n_,
			FMTSymbols`ifT[_,then_]
		]
		:>
		List[
			pathStepT[ToString[n]<>". If "<>" ..."],
			pathBranchesT[then,{}]
		],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[FMTSymbols`ifT[_,then_]]
		]
		:>
		List[
			pathStepT[Highlighted[ToString[n]<>". If "<>" ..."]],
			pathBranchesT[then,{}]
		],
		
		(* ifT with both (then/else) branches: *)
		
		FMTSymbols`numberedT[n_,
			FMTSymbols`ifT[_,then_,else_]
		]
		:>
		List[
			pathStepT[ToString[n]<>". If "<>" ..."],
			pathBranchesT[then,else]
		],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[FMTSymbols`ifT[_,then_,else_]]
		]
		:>
		List[
			pathStepT[Highlighted[ToString[n]<>". If "<>" ..."]],
			pathBranchesT[then,else]
		],
		
		(* whileT: *)
		
		FMTSymbols`numberedT[n_,
			FMTSymbols`whileT[_,loopBody_]
		]
		:>
		List[
			pathStepT[ToString[n]<>". While "<>" ..."],
			pathBranchesT[loopPathT[loopBody],{}]
		],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[FMTSymbols`whileT[_,loopBody_]]
		]
		:>
		List[
			pathStepT[Highlighted[ToString[n]<>". While "<>" ..."]],
			pathBranchesT[loopPathT[loopBody],{}]
		],
		
		(* switchT: *)
		
		FMTSymbols`numberedT[n_,
			FMTSymbols`switchT[FMTSymbols`objectT[condObj_,_],cases__]
		]
		:>
		List[
			pathStepT[ToString[n]<> ". Switch "<>ToString[condObj]],
			pathBranchesT[cases][[2;; ;;2]] (* process only even elements, odd elements are case labels *)
		],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[FMTSymbols`switchT[FMTSymbols`objectT[condObj_,_],cases__]]
		]
		:>
		List[
			pathStepT[Highlighted[ToString[n]<> ". Switch "<>ToString[condObj]]],
			pathBranchesT[cases][[2;; ;;2]] (* process only even elements, odd elements are case labels *)
		],
		
		(* operation calls: *)
		
		FMTSymbols`numberedT[n_,
			FMTSymbols`callT[qOpName_,_,_]
		]
		:>
		pathStepT[ToString[n]<>". "<>ToString[qOpName]<>"[...]"],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[FMTSymbols`callT[qOpName_,_,_]]
		]
		:>
		pathStepT[Highlighted[ToString[n]<>". "<>ToString[qOpName]<>"[...]"]],
		
		(* assignments: *)
		
		FMTSymbols`numberedT[n_,
			FMTSymbols`assignmentT[FMTSymbols`objectLhsT[obj_,_],_]
		]
		:>
		pathStepT[ToString[n]<>". "<>shortObjectName[obj]<>" := ..."],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[FMTSymbols`assignmentT[FMTSymbols`objectLhsT[obj_,_],_]]
		]
		:>
		pathStepT[Highlighted[ToString[n]<>". "<>ToString[obj]<>" := ..."]],
		
		(* ghost captures: *)
		
		FMTSymbols`numberedT[n_,
			FMTSymbols`ghostCaptureT[name_String,_]
		]
		:>
		pathStepT[ToString[n]<>". ghost "<>name<>" == ..."],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[FMTSymbols`ghostCaptureT[name_String,_]]
		]
		:>
		pathStepT[Highlighted[ToString[n]<>". ghost "<>name<>" == ..."]],
		
		(* assertions and assumptions: *)
		
		FMTSymbols`numberedT[n_,
			_FMTSymbols`assertT
		]
		:>
		pathStepT[ToString[n]<>". Assert[...]"],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[_FMTSymbols`assertT]
		]
		:>
		pathStepT[Highlighted[ToString[n]<>". Assert[...]"]],
		
		FMTSymbols`numberedT[n_,
			_FMTSymbols`assumeT
		]
		:>
		pathStepT[ToString[n]<>". Assume[...]"],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[_FMTSymbols`assumeT]
		]
		:>
		pathStepT[Highlighted[ToString[n]<>". Assume[...]"]],
		
		FMTSymbols`numberedT[n_,
			_FMTSymbols`deduceT
		]
		:>
		pathStepT[ToString[n]<>". Deduce[...]"],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[_FMTSymbols`deduceT]
		]
		:>
		pathStepT[Highlighted[ToString[n]<>". Deduce[...]"]],
		
		FMTSymbols`numberedT[n_,
			_FMTSymbols`induceT
		]
		:>
		pathStepT[ToString[n]<>". Induce[...]"],
		(* highlighted version: *)
		FMTSymbols`numberedT[n_,
			FMTSymbols`highlightedT[_FMTSymbols`induceT]
		]
		:>
		pathStepT[Highlighted[ToString[n]<>". Induce[...]"]]
	};
	
	expression/.replacements
]


(* Helper, optionally highlights the given point of the flow diagram
	(used only for "start" and "finish" points, the rest are highlighed by the grammar resolver)
*)
highlightEndPoint[point_,highlightedSet_List]:=
	If[MemberQ[highlightedSet,point],
		Highlighted[point],
		point
	]


(* Helper, removes the given vertex from the graph and short-cuts all adjacent edges,
	according to the edge directions (that is, preserve all possible flows) *)
bypassNode[edges_,v_]:=Module[
	{
		rest
	},
	
	rest=DeleteCases[edges,_\[DirectedEdge]v];
	rest=DeleteCases[rest,v\[DirectedEdge]_];
	
	Join[
		rest,
		DirectedEdge@@@Tuples[
			{
				Cases[edges,from_\[DirectedEdge]v->from],
				Cases[edges,v\[DirectedEdge]to_->to]
			}
		]
	]
]


(* Helper, builds a graph from expression (in a simplified control flow form)
	and adds given starting and finishing points *)

buildGraph[from_,to_,{}]:={from\[DirectedEdge]to}

buildGraph[from_,to_,step_pathStepT]:=With[
	{
		stepContent=Identity@@step (* drop head *)
	},
	
	{from\[DirectedEdge]stepContent,stepContent\[DirectedEdge]to}
]

buildGraph[from_,to_,steps_List]:=Module[
	{
		flatSteps,
		path,
		branchPosition,
		branchFrom, (* placeholder symbol to mark the place where nested construct begins *)
		branchTo, (* placeholder, where the nested construct ends *)
		loopReturnPoint=None, (* placeholder for the end of looping path *)
		branchPaths
	},
	
	flatSteps=Flatten[steps];
	
	If[MatchQ[flatSteps,List[__pathStepT]],
		(* flatSteps is {pathStepT[a],pathStepT[b],pathStepT[c],...}
			that is, flatSteps are not nested and there are no branches,
			this is a simple sequence of terminal pathStepT elements *)
		With[
			{
				(* extract all step contents: drop inner heads *)
				stepContents=Identity@@@flatSteps
			},
			
			(* turn {a,b,c,...} into {from\[DirectedEdge]a,a\[DirectedEdge]b,b\[DirectedEdge]c,...\[DirectedEdge]to} *)
			path=DirectedEdge@@@Partition[
				Flatten[{from,stepContents,to}],
				2,1
			]
		],
		
		(* otherwise, flatSteps contains branches *)
		
		(* "cut" the nested construct out from the path and process the parts separately *)
		(* Note: only one branched construct is cut out here; others, if present,
			are handled in recursive calls for the remaining parts of the whole expression
		*)
		
		branchPosition=Position[flatSteps,_pathBranchesT,{1},1];
		If[branchPosition!={},
			(* take only the first of all pathBranchesT components: *)
			branchPosition=branchPosition[[1,1]];
			
			branchFrom=Unique[];
			branchTo=Unique[];
			
			(* process recursively all branched paths: *)
			branchPaths=Map[
				Function[branch,
					If[Head[branch]===loopPathT,
						(* this branch-path is wrapped in loopPathT,
							which designates that the path should finish with a jump
							to the place that precedes the path
							(return to the beginning of the loop);
							Note: loopPathT never exists as the first element in the list of steps,
								there is always something before (If/While condition),
								so that closing the cycle to branchPosition-1 is always safe.
						*)
						loopReturnPoint=Unique[];
						{
							buildGraph[branchFrom,loopReturnPoint,branch[[1]]],
							buildGraph[
								loopReturnPoint,
								flatSteps[[branchPosition-1]][[1]], (* preceding entry *)
								{}
							]
						},
						
						buildGraph[branchFrom,branchTo,branch]						
					]
				],
				List@@flatSteps[[branchPosition]]
			];
			
			path=Join[
				buildGraph[from,branchFrom,flatSteps[[;;branchPosition-1]]],
				Flatten[branchPaths],
				buildGraph[branchTo,to,flatSteps[[branchPosition+1;;]]]
			]
		];
		
		(* remove artificial cut points: *)
		
		path=bypassNode[path,branchFrom];
		path=bypassNode[path,branchTo];
		If[loopReturnPoint=!=None,
			path=bypassNode[path,loopReturnPoint]
		]
	];
	
	path
]


controlFlow[expression_,highlightedSet_List]:=Module[
	{
		simplified,
		pathGraph
	},
	
	simplified=FixedPoint[controlFlowReduceStep,expression];
	If[simplified===Nothing, (* special case for empty procedure body *)
		simplified={}
	];
	
	pathGraph=buildGraph[
		highlightEndPoint["start",highlightedSet],
		highlightEndPoint["finish",highlightedSet],
		simplified
	];

	(* finally, reverse the loopStepT segments by creating reverse direct edges from them;
		this needs to be done after the call to buildGraph, which is recursive
		and reversing loopStepT segments there conflicts with outer recursion calls
	*)	
	pathGraph/.loopStepT[from_,to_]->DirectedEdge[to,from]
]


dataFlow[expression_,paramModes_]:=Module[
	{
		inParams,
		assignments,
		calls,
		result={}
	},
	
	inParams=Select[paramModes,
		Function[p, (* p is {name,mode} *)
			With[
				{
					mode=p[[2]]
				},
				
				mode===FMTSymbols`in
			]
		]
	][[All,1]];
	
	assignments=Cases[expression,
		FMTSymbols`numberedT[index_,FMTSymbols`assignmentT[FMTSymbols`objectLhsT[name_,_],_]]
			->{index,name},
		All
	];
	
	(* Note: some of the objects assigned to might be subojbects of respective parent objects
		(like elements of complete arrays or fields of structures),
		these are identified below, in addition to the main left-hand-side object
	*)
	Scan[
		Function[as, (* as is {index,objName} *)
			With[
				{
					index=as[[1]],
					objName=as[[2]],
					
					(* potential parents in lhs: *)
					parentObjNames=Cases[as[[2]],FMTSymbols`objectParentT[name_,_]->name,All]
				},
				
				Scan[
					Function[obj, (* obj is either objName or any objectParentT found *)
						If[MemberQ[inParams,obj],
							AppendTo[result,
								{index,"Parameter '"<>ToString[obj]<>"' cannot be assigned to."}
							]
						]
					],
					Join[{objName},parentObjNames]
				]
			]
		],
		assignments
	];
	
	calls=Cases[expression,
		FMTSymbols`numberedT[index_,call_FMTSymbols`callT]->{index,call},
		All
	];
	
	Scan[
		Function[c, (* c is {index,callT[qOpName,formals,actuals]} *)
			With[
				{
					index=c[[1]],
					formals=List@@c[[2,2]],
					actuals=List@@c[[2,3]]
				},
				
				Scan[
					Function[pp, (* pp is {formalParam,actualParam},
									where formalParam is mode[name\[Rule]type]
									and actualParam is objectT[name,type] *)
						With[
							{
								formMode=Head[pp[[1]]],
								actName=pp[[2,1]]
							},
							
							If[(formMode===FMTSymbols`inOut||
								formMode===FMTSymbols`out)&&
								MemberQ[inParams,actName],
								
								AppendTo[result,
									{
										index,
										"Parameter '"<>ToString[actName]<>
										"' cannot be used in a call as an inOut or out actual."
									}
								]
							]
						]
					],
					Transpose[{formals,actuals}]
				]
			]
		],
		calls
	];
	
	result
]


outgoingCalls[expression_]:=
	Cases[expression,FMTSymbols`callT[qOpName_,formals_,actuals_]->qOpName,All]


indirectObjectUses[directUses_Association,operationCalls_Association]:=
	FixedPoint[
		Function[uses, (* uses is <|opName\[Rule]{obj1,obj2,...},...|> *)
			Module[{usesClosure=uses},
				Scan[
					Function[singleOpCalls, (* singleOpCalls is opName\[Rule]{op1,op2,...} *)
						With[
							{
								caller=singleOpCalls[[1]],
								calledOperations=singleOpCalls[[2]]
							},
							
							Scan[
								Function[calledOp,
									(* Note: the analysis might be performed for a subset
										of a complete system, which can lead to a situation
										where some called operation (as found in the body
										of some operation from the analysed part) has no
										information about its used objects (because the called
										operation itself was not analysed)
										- assume empty set of used objects in this case.
									*)
									usesClosure[caller]=Union[
										usesClosure[caller],
										Lookup[uses,calledOp,{}] (* assume empty for unchecked callees *)
									]
								],
								calledOperations
							]
						]
					],
					Normal[operationCalls]
				];
				
				usesClosure
			]
		],
		directUses
	]


dataAliasing[qOperationName_Symbol,expression_,globalsUsed_Association,globalsIndirectlyUsed_Association]:=Module[
	{
		calls,
		index,
		qCalledOp,
		formalModes,
		actualNames,
		actualModes,
		callerGlobals,
		calleeGlobals,
		singleOpAliases={}
	},
	
	calls=Cases[expression,
		FMTSymbols`numberedT[index_,c_FMTSymbols`callT]
			->{index,c},
		All
	];
	
	Scan[
		Function[c, (* c is {index,callT[qCalledOp,formals[...],actuals[...]]} *)
			index=c[[1]];
			qCalledOp=c[[2,1]];
			formalModes=List@@Head/@c[[2,2]];
			actualNames=List@@c[[2,3,All,1]];
			
			(* 1. collect aliasings in operation parameters *)
			
			(* actualModes is <|name\[Rule]{mode1,mode2,...},...|>
				for all actual parameters *)
			actualModes=Merge[Thread[actualNames->formalModes],Identity];
			
			Scan[
				Function[a, (* a is name\[Rule]{mode1,...} *)
					With[
						{
							aName=a[[1]],
							modes=a[[2]]
						},
						
						(* find actuals that are used more than once
							and at least one of the use is of inOut/out mode *)
						If[Length[modes]>1&&
							MemberQ[modes,FMTSymbols`inOut|FMTSymbols`out],
							
							AppendTo[singleOpAliases,
								{
									index,
									"Object '"<>ToString[aName]<>"' is aliased in this call."
								}
							]
						]
					]
				],
				Normal[actualModes]
			];
			
			(* 2. collect aliasing in global objects
				- that is, when caller and callee (including nested) refer to the same global;
				the case of unrelated operations using the same global is not taken into account,
				because it is not, as such, a hidden aliasing, only nested uses are collected here *)
			
			callerGlobals=globalsUsed[qOperationName];
			calleeGlobals=Lookup[globalsIndirectlyUsed,qCalledOp,{}];
			Scan[
				Function[gName, (* gName is aliased global data object *)
					AppendTo[singleOpAliases,
						{
							index,
							"Global object '"<>ToString[gName]<>"' is internally used by the called operation."
						}
					]
				],
				Intersection[callerGlobals,calleeGlobals]
			]
		],
		calls
	];
	
	singleOpAliases
]


End[];


EndPackage[];
