(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


BeginPackage["FMTCodeFormatter`"];


format::usage="format[tokens,options] Compiles the given list of tokens, which can contain strings, wrappers and formatter symbols,\n"<>
"returns the formatted string. Options is an association of options that can affect formatting.";


formatterOptionNames={
	indentSize, (* number of spaces used for indentation (mini-indent is hardwired to be 1 space less) *)
	maxLineLength (* max length after which (softly) the line is automatically broken *)
};

codeFormatterSymbols={
	moreIndentT, (* increases indent for subsequent lines *)
	lessIndentT, (* reduces indent *)
	verticalSpaceT, (* ensures there is a vertical white space before further text *)
	continueWithVerticalSpaceT, (* inserts a vertical space only if no change in indent level *)
	lineBreakT, (* inserts line break *)
	noLineBreakT, (* wrapper, ensures there is no line break between this and following noLineBreadT token *)
	miniIndentIfContinuedT, (* wrapper, ensures that, if there was a need to break the line, the continuation will start with a mini-indent *)
	canTrimT (* wrapper, allows right-trimming of the contained whitespace, if it appears at the beginning or end of the line *)
};


Begin["`Private`"];


(* Helper, allows 0-times repeats *)
strRepeat[str_String,n_Integer]:=If[n>0,StringRepeat[str,n],""]

format[tokens_List,options_Association]:=Module[
	{
		trimmedTokens,
		maxLineLength=options[maxLineLength],
		currentLineLength=0,
		currentIndentLevel=0,
		indent=strRepeat[" ",options[indentSize]],
		miniIndent=strRepeat[" ",options[indentSize]-1],
		pendingIndent="",
		atBeginOfLine=True,
		verticalSpaceAlreadyDone=True, (* treat beginning of text as a new paragraph *)
		result={}
	},
	
	(* remove trailing vertical space tokens *)
	trimmedTokens=tokens/.List[pre__,v__verticalSpaceT]->List[pre];
	
	Scan[
		Function[token,
			Switch[Head[token],
				String,
					(* regular string content *)
					If[currentLineLength>=maxLineLength,
						AppendTo[result,"\n"];
						currentLineLength=0;
						
						(* by default use deeper indent for line continuations;
							this indent can be changed by subsequent miniIndentIfContinuedT wrapper
						*)
						pendingIndent=strRepeat[indent,currentIndentLevel+1]
					];
					AppendTo[result,pendingIndent];
					currentLineLength+=StringLength[pendingIndent];
					pendingIndent="";
					AppendTo[result,token];
					atBeginOfLine=False;
					verticalSpaceAlreadyDone=False;
					currentLineLength+=StringLength[token];
					If[currentLineLength>=maxLineLength,
						AppendTo[result,"\n"];
						atBeginOfLine=True;
						currentLineLength=0;
						
						(* by default use deeper indent for line continuations;
							this indent can be changed by subsequent miniIndentIfContinuedT wrapper
						*)
						pendingIndent=strRepeat[indent,currentIndentLevel+1]
					],
				
				moreIndentT,
					++currentIndentLevel;
					pendingIndent=strRepeat[indent,currentIndentLevel],
				
				lessIndentT,
					--currentIndentLevel;
					pendingIndent=strRepeat[indent,currentIndentLevel],
				
				verticalSpaceT,
					(* reset indenting *)
					pendingIndent=strRepeat[indent,currentIndentLevel];
					If[atBeginOfLine,
						If[Not[verticalSpaceAlreadyDone],
							(* if the cursor is already at the beginning of line,
								one newline is enough to make a vertical white space *)
							AppendTo[result,pendingIndent]; (* keep indenting even in empty lines *)
							AppendTo[result,"\n"];
							verticalSpaceAlreadyDone=True
						],
						
						AppendTo[result,"\n"];
						AppendTo[result,pendingIndent];
						AppendTo[result,"\n"];
						atBeginOfLine=True;
						currentLineLength=0;
						verticalSpaceAlreadyDone=True
					],
				
				continueWithVerticalSpaceT,
					pendingIndent=pendingIndent<>"\n"<>pendingIndent,
				
				lineBreakT,
					If[Not[atBeginOfLine],
						AppendTo[result,"\n"]
					];
					pendingIndent=strRepeat[indent,currentIndentLevel];
					atBeginOfLine=True;
					currentLineLength=0;
					verticalSpaceAlreadyDone=False,
				
				noLineBreakT,
					(* content wrapped in noLineBreakT is treated as regular String,
						with the exception that lines are not automatically broken
						if such a token crosses the maxLineLength limit
					*)
					AppendTo[result,pendingIndent];
					currentLineLength+=StringLength[pendingIndent];
					pendingIndent="";
					AppendTo[result,token[[1]]];
					atBeginOfLine=False;
					verticalSpaceAlreadyDone=False;
					currentLineLength+=StringLength[token[[1]]],
				
				miniIndentIfContinuedT,
					(* miniIndent overrides the existing pending indent *)
					If[currentLineLength>=maxLineLength||pendingIndent!="",
						pendingIndent=strRepeat[indent,currentIndentLevel]<>miniIndent
					];
					(* the wrapped content is generally processed as a regular token,
						which means that it breaks the line if it is already past the limit,
						but not if it crosses the limit after printed
					*)
					If[currentLineLength>=maxLineLength,
						AppendTo[result,"\n"];
						currentLineLength=0
					];
					AppendTo[result,pendingIndent];
					currentLineLength+=StringLength[pendingIndent];
					pendingIndent="";
					AppendTo[result,token[[1]]];
					atBeginOfLine=False;
					verticalSpaceAlreadyDone=False;
					currentLineLength+=StringLength[token[[1]]],
				
				canTrimT,
					(* content wrapped in canTrimT can be trimmed,
						or just ommitted, because it is a whitespace anyway,
						if it happens at the end of the line,
						otherwise it is processed as a regular token
					*)
					If[Not[atBeginOfLine],
						If[currentLineLength+StringLength[token[[1]]]>=maxLineLength,
							AppendTo[result,"\n"];
							atBeginOfLine=True;
							currentLineLength=0;
							verticalSpaceAlreadyDone=False;
							
							(* by default use deeper indent for line continuations;
								this indent can be changed by subsequent miniIndentIfContinuedT wrapper
							*)
							pendingIndent=strRepeat[indent,currentIndentLevel+1],
							
							AppendTo[result,token[[1]]];
							atBeginOfLine=False;
							verticalSpaceAlreadyDone=False;
							currentLineLength+=StringLength[token[[1]]]
						]
					]
			]
		],
		trimmedTokens
	];
	
	StringJoin[result]
]


End[];


EndPackage[];
