(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


(*Needs["FMT`"]; - circular dependency, acceptable here *)
Needs["FMTSymbols`"];


BeginPackage["FMTFormatters`"];


formatNameChecks::usage="formatNameChecks[result] Formats results of name conflict checks.";


formatGrammarChecks::usage="formatGrammarChecks[system,result] Formats results of grammar checks.";


formatGrammarReductions::usage="formatGrammarReductions[reductions] Formats list of grammar reductions as a column of tree plots.";


formatGrammar::usage="formatGrammar[expression] Format the given expression.";


formatFlowChecks::usage="formatFlowChecks[system,result] Formats results of flow checks.";


formatDataAliasingChecks::usage="formatDataAliasingChecks[system,result] Formats results of data aliasing checks.";


formatCallTable::usage="formatCallTable[result] Formats tabular results of operation call analysis.";


formatCallGraph::usage="formatCallGraph[result,entryPoint] Formats results of call analysis as a graph.";


formatCallCycles::usage="formatCallCycles[result,entryPoint] Formats results of call cycle analysis.";


formatControlFlow::usage="formatControlFlow[flow] Formats the control flow structure as a graph.";


formatProofStatements::usage="formatProofStatements[system,packageName,operationName,proofStatements,formulaFormatter] Formats the list of proof statements as a grid.";


formatVariableDomains::usage="formatVariableDomains[domains] Formats the list of variable domains as a grid.";


formatSingleOpProofChecks::usage="formatSingleOpProofChecks[system,packageName,operationName,results,formulaFormatter] Formats results of proof checks for a single operation.";


formatProofChecks::usage="formatProofChecks[system,result,formulaFormatter] Formats results of proof checks for multiple operations.";


Begin["`Private`"];


(* Helper for common grid style properties *)
commonGridStyle=Sequence[
	Spacings->{1,1},
	Frame->All,
	Background->{None,1->LightBlue} (* Header only *)
];


(* Helper for package/operation formatters,
	creates a grid with results for each operation name, grouped by package name;
	thirdTitle is a title for the third column (result) of the generated result table,
	thirdFormatter is a function that ultimately prepares the presentation form
		for the single result *)
genericFormat[report_,thirdTitle_String,thirdFormatter_]:=Module[
	{
		temp
	},
	
	Grid[
		Join[
			{{"Package","Operation",thirdTitle}},
			Flatten[
				Map[
					Function[outerKey,
						temp=Map[
							Function[innerKey,
								{
									SpanFromAbove,
									innerKey,
									thirdFormatter[outerKey,innerKey,report[outerKey][innerKey]]
								}
							],
							Keys[report[outerKey]]
						];
						
						If[temp!={},
							temp[[1,1]]=outerKey,
							temp=Nothing
						];
						
						temp
					],
					Keys[report]
				],
				1
			]
		],
		Alignment->{Left,Top}, (* Baseline instead of Top?
									Baseline looks good with single lines in all columns,
									but a bit worse with Column on the right. *)
		commonGridStyle
	]
]


formatNameChecks[result_]:=Column[{
	"Package-level conflicts:",
	Grid[
		Join[
			{{"Package","Conflicts"}},
			Map[
				Function[pkgName,
					With[
						{
							pkgConflicts=result[pkgName][[1]]
						},
						
						{
							pkgName,
							If[pkgConflicts=={},
								Style["OK",Green],
								Style[Column[pkgConflicts],Red]
							]
						}
					]
				],
				Keys[result]
			]
		],
		Alignment->{Left,Baseline},
		commonGridStyle
	],
	"",
	"Operation-level conflicts:",
	genericFormat[result[[All,2]],"Conflicts",
		Function[{pkgName,opName,singleResult},
			If[singleResult=={},
				Style["OK",Green],
				
				Style[Column[singleResult],Red]
			]
		]
	]
}]


formatGrammarChecks[system_,result_]:=genericFormat[result,"Grammar",
	Function[{pkgName,opName,singleResult},
		If[singleResult,
			Style["OK",Green],
			
			(* if the result for a single operation is False,
				give the user possibility to display the reduction tree
				for the failing operation
			*)
			Row[{
				Style["Failed  ",Red],
				Button["show tree",
					CreateDocument[
						{
							TextCell[
								"Grammar tree for "<>pkgName<>ToString[opName],
								"Subsection"
							],
							ExpressionCell[
								FMT`showGrammarReductions[system,pkgName,opName],
								"Output"
							]
						}
					]
				]
			}]
		]
	]
]


formatGrammarReductions[reductions_]:=Column[
	TreeForm/@reductions
]


formatGrammar[expression_]:=TreeForm[expression]


(* Helper for formatting failure reports
	that have a "show" button displayed in a row with failure description
	in the right column of the grid *)
simpleFailureDescriptionFormatter[system_]:=
	Function[{pkgName,opName,singleResult},
		If[singleResult=={},
			Style["OK",Green],
			
			(* if the result for a single operation contains failures,
				give the user possibility to display the control flow diagrams
				for the offending index
			*)
			With[
				{
					singleFormatted=formatIndexColumn[
						system,pkgName,opName,singleResult,Row," "]
				},
				
				Column@Map[
					Function[failure, (* failure is {index,description} *)
						With[
							{
								index=failure[[1]],
								description=failure[[2]]
							},
							
							Row[{
								index,
								" ",
								Style[description,Red,Italic]
							}]
						]
					],
					singleFormatted
				]
			]
		]
	]


formatFlowChecks[system_,result_]:=
	genericFormat[result,"Flows",simpleFailureDescriptionFormatter[system]]


formatDataAliasingChecks[system_,result_]:=
	genericFormat[result,"Data aliasing",simpleFailureDescriptionFormatter[system]]


(* Helper, creates short labels with tooltips for all vertices used in the call graph *)
callLabels[result_]:=With[
	{
		vertices=Union[Flatten[List@@@result]]
	},
	
	Map[
		Function[v, (* v is a single vertex name *)
			With[
				{
					nameComponents=StringSplit[ToString[v],"`"]
				},
				
				v->Tooltip[Last[nameComponents],v]
			]
		],
		vertices
	]
]

formatCallTable[result_]:=
	genericFormat[result,"Calls",
		Function[{pkgName,opName,singleResult},
			Style[Column[singleResult[[All,2]]],Blue]
		]
	]

formatCallGraph[result_,entryPoint_]:=
	If[result!={},
		Graph[
			result,
			GraphLayout->If[entryPoint=!=None,
				{"LayeredDigraphEmbedding","RootVertex"->entryPoint},
				Automatic
			],
			VertexLabels->callLabels[result]
		],
		Style["No operation calls found.",Italic]
	]


formatCallCycles[result_,entryPoint_]:=With[
	{
		callGraph=result[[1]],
		cycles=result[[2]]
	},
	
	If[cycles=={},
		Style["No call cycles.",Green],
		
		Column@Flatten[{
			Style["Detected call cycles:",Red],
			Map[
				Function[c, (* c is a cycle *)
					HighlightGraph[formatCallGraph[callGraph,entryPoint],c,GraphHighlightStyle->"Thick"]
				],
				cycles
			]
		},
		1
		]
	]
]


formatControlFlow[flow_]:=Graph[flow,
	GraphLayout->{"LayeredDigraphEmbedding","RootVertex"->flow[[1,1]]},
	VertexLabels->Automatic
]


(* Helpers *)

(* transforms initializedT into superscript: *)
formatInitialized[expr_]:=
	expr/.FMTSymbols`initializedT[name_]->Superscript[name,"init"]

(* adds "show" buttons to indices in first column: *)
formatIndexColumn[system_,packageName_,operationName_,expr_,arranger_,spacer_]:=
	If[expr!={},
		MapAt[
			Function[index,
				arranger[
					{
						index,
						spacer,
						Button["show",
							CreateDocument[
								{
									TextCell[
										"Control flow diagram for "<>packageName<>ToString[operationName],
										"Subsection"
									],
									ExpressionCell[
										formatControlFlow[
											FMT`reportControlFlow[system,packageName,operationName,highlighted->{index}]
										],
										"Output"
									]
								}
							]
						]
					},
					Alignment->If[arranger===Column,Center,Baseline]
				]
			],
			expr,
			{All,1}
		],
		{}
	]

(* shows description in italics and colors the alwaysT/possibleT/neverT tags: *)
formatSingleVerificationStatement[verSt_,formulaFormatter_]:=With[
	{
		description=verSt[[1]],
		statementHead=verSt[[2,0]],
		statementBody=Column[List@@verSt[[2]]]
	},
	
	{
		Style[description,Italic],
		Row[{
			Switch[statementHead,
				FMTSymbols`alwaysT,
					Style["always: ",Green,Italic],
				
				FMTSymbols`possibleT,
					Style["possible: ",Blue,Italic],
				
				FMTSymbols`neverT,
					Style["never: ",Red,Italic],
				
				FMTSymbols`deductionT,
					Style["deduce: ",Purple,Italic],
				
				FMTSymbols`inductionT,
					Style["induce: ",Purple,Italic]
			],
			formulaFormatter[statementBody]
		}]
	}
]



formatProofStatements[system_,packageName_String,operationName_Symbol,proofStatements_,formulaFormatter_]:=Module[
	{
		statementsFormatted={}
	},
	
	If[proofStatements!={},
		(* 1. format initializedT wrappers as superscripts *)
		
		statementsFormatted=formatInitialized[proofStatements];
		
		(* 2. format first column "Index" to enable links to control flow diagrams
			with the given index highlighted *)
		
		statementsFormatted=formatIndexColumn[
			system,packageName,operationName,statementsFormatted,Column,Nothing];
		
		(* 3. formst "Knowns" according to the selected format *)
		statementsFormatted=MapAt[
			Function[knowns,
				formulaFormatter/@knowns
			],
			statementsFormatted,
			{All,2}
		];
		
		(* 4. format third "Verification" column for nice display *)
		
		statementsFormatted=MapAt[
			Function[verSts, (* verSts is {{description,statement},...} *)
				Flatten[
					Riffle[
						Map[
							Function[verSt, (* verSt is a single described verification statement:
												{description,statement} *)
								formatSingleVerificationStatement[verSt,formulaFormatter]
							],
							verSts
						],
						"\n" (* add spacing between verification statements *)
					]
				]
			],
			statementsFormatted,
			{All,3}
		];
		
		(* 4. format everything for proper display (Column) of list content,
			note: first column is already formatted: *)
		
		statementsFormatted=MapAt[
			Column,
			statementsFormatted,
			{All,2;;3}
		]
	];
	
	If[statementsFormatted!={},
		Grid[
			(* extend the list of proof statements by a grid header: *)
			Join[
				{{"Index","Knowns","Verification"}},
				statementsFormatted
			],
			Alignment->{{Center,Left,Left},Top},
			commonGridStyle
		],
		
		Style["There are no proof statements.",Italic]
	]
]


formatVariableDomains[domains_]:=Grid[
	List@@@Normal[domains],
	Frame->All
]


formatSingleOpProofChecks[system_,packageName_String,operationName_Symbol,results_,formulaFormatter_]:=Module[
	{
		resultsFormatted={}
	},
	
	If[results!={},
		(* 1. format initializedT wrappers as superscripts *)
		
		resultsFormatted=formatInitialized[results];
		
		(* 2. format first column "Index" to enable links to control flow diagrams
			with the given index highlighted *)
		
		resultsFormatted=formatIndexColumn[
			system,packageName,operationName,resultsFormatted,Column,Nothing];
		
		(* 3. format second "Verification" column for nice display *)
		
		resultsFormatted=MapAt[
			Function[verSt, (* verSt is a single described verification statement:
								{description,statement} *)
				Column[formatSingleVerificationStatement[verSt,formulaFormatter]]
			],
			resultsFormatted,
			{All,2}
		];
		
		(* 4. format the 3rd result column *)
		
		resultsFormatted=MapAt[
			Function[res,
				Switch[Head[res],
					FMTSymbols`provenT,
						Style["OK",Green],
					
					FMTSymbols`impossibleT,
						Style["Impossible",Red],
					
					FMTSymbols`failedT,
						Column[{
							Style["Failed",Red],
							If[Length[res]>0,
								If[Length[res]==1&&Head[res[[1,1]]]===String,
									(* special case for single-line textual descriptions: *)
									Style[res[[1,1]]<>": ",res[[1,2]],Black],
									
									(* make a tooltip with the list of counter-examples *)
									Tooltip[Highlighted["see counter-examples"],
										Column[
											List@@res,
											Spacings->1
										]
									]
								],
								Style["No failure context found.",Italic]
							]
						}]
				]
			],
			resultsFormatted,
			{All,3}
		]
	];
	
	If[resultsFormatted!={},
		Grid[
			(* extend the list of proof statements by a grid header: *)
			Join[
				{{"Index","Verification","Result"}},
				resultsFormatted
			],
			Alignment->{{Center,Left,Left},Top},
			commonGridStyle
		],
		
		Style["No proof results to show.",Italic]
	]
]


formatProofChecks[system_,result_,formulaFormatter_]:=genericFormat[result,"Proofs",
	Function[{pkgName,opName,singleResult},
		With[
			{
				consolidatedResult=Union[Head/@singleResult[[All,3]]]
			},
			
			If[consolidatedResult==={FMTSymbols`provenT}||consolidatedResult==={},
				Style["OK",Green],
				
				(* if the result for a single operation is Failed
					(that is, it contains entries other than provenT),
					give the user possibility to display the detailed report
					for the failing operation
				*)
				Row[{
					Style["Failed  ",Red],
					Button["show proofs",
						CreateDocument[
							{
								TextCell[
									"Proof results for "<>pkgName<>ToString[opName],
									"Subsection"
								],
								ExpressionCell[
									formatSingleOpProofChecks[system,pkgName,opName,singleResult,formulaFormatter],
									"Output"
								]
							}
						]
					],
					" ",
					Button["show statements",
						CreateDocument[
							{
								TextCell[
									"Proof statements for "<>pkgName<>ToString[opName],
									"Subsection"
								],
								ExpressionCell[
									FMT`showProofStatements[system,pkgName,opName],
									"Output"
								]
							}
						]
					]
				}]
			]
		]
	]
]


End[];


EndPackage[];
