(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTOptions`"];
Needs["FMTSymbols`"];


BeginPackage["FMTGrammar`"];


resolveSyntax::usage="resolveSyntax[expression,preprocessor,typeHier,enumValues,visibleObjects,visibleEnums,visibleOperations,numbered,highlighted] Replaces literals, symbols and standard operations with wrapped or safe forms for further processing.\n"<>
"The preprocessor is a function that performs arbitrary transformation of the given expression at the beginning of processing.\n"<>
"The typeHier is a list of rules defining all types in the system,\n"<>
"enumValues is an association with value definitions for all enumeration types.";
"The visibleObjects argument has the form {objectName\[Rule]{fullyQualifiedName,typeTag},...}.\n"<>
"The visibleEnums list has the form {enumName->{fullyQualifiedName,typeTag},...}.\n"<>
"The visibleOperations has the form {opName->qOpName[...],...}.\n"<>
"The numbered argument is an option, by default False, specifying whether statements should be wrapped and numbered.\n"<>
"The highlighted argument is an option, taking a list of numbers defining which of the numbered parts of the expression should be wrapped in the highlightedT wrapper.";


unwrapGrammar::usage="unwrapGrammar[expression,options] Reverts the grammar resolution into human-readable forms.\n"<>
"Note: this function is intended for wrapped objects and single expressions, not for control structures or complete operation bodies.";


accessTopObject::usage="accessTopObject[object] Identifies the top-level object name in dotT and partT constructs.";


reduceAll::usage="reduceAll[typeHier,enumValues,resolvedExpr] Returns the final result of grammar reduction for the given expression,\n"<>
"type hierarchy and a set of enum definitions.";


grammarCorrectQ::usage="grammarCorrectQ[typeHier,enumValues,resolvedExpr] Returns True if the given (already resolved) expression conforms to grammar."<>
"Otherwise, reductionList can reveal the non-conforming parts.\n"<>
"The typeHier is a list of rules defining type hierarchy for those types that are based on other types\n"<>
"enumValues is an association with value definitions for all enumeration types.";


reductionList::usage="reductionList[typeHier,enumValues,resolvedExpr] Returns the list of consecutive expression reductions,"<>
" starting from the given (already resolved) expression.\n"<>
"The typeHier is a list of rules defining type hierarchy for those types that are based on other types,\n"<>
"enumValues is an association with value definitions for all enumeration types.";


Begin["`Private`"];


(* Helper, wraps literals into typed expression parts (integerT, enums, etc.) *)
sanitizeLiterals[expression_,visibleEnums_List]:=Module[
	{
		e
	},
	
	e=Replace[expression,v_Integer->FMTSymbols`literalT[v,FMTSymbols`integerT],{-1}];
	e=Replace[e,True->FMTSymbols`literalT[True,FMTSymbols`booleanT],{-1}];
	e=Replace[e,False->FMTSymbols`literalT[False,FMTSymbols`booleanT],{-1}];
	e=Replace[e,Blank->FMTSymbols`defaultT,{-1},Heads->True];
	
	(* sanitize enums as well, according to their visibility and lookup rules *)
	e=e/.Map[
		Function[el, (* el is name->{qName,qType} *)
			With[
				{
					eName=el[[1]],
					eQName=el[[2,1]],
					eQType=el[[2,2]]
				},
				
				eName->FMTSymbols`literalT[eQName,eQType]
			]
		],
		visibleEnums
	];
	
	e
]


(* Helper, wraps symbols into objectT references,
	except those that appear directly under dotT wrappers *)
(* visibleObjects has the form {s\[Rule]{fullyQualifiedName,type},...} *)
resolveObjects[expression_,visibleObjects_List]:=Module[
	{
		temp,
		replacements
	},
	
	(* structure field names are not wrapped as stand-alone objects,
		in order to omit them during this process, especially when
		there is a name clash if the field name is the same as one of the
		existing object, the following strategy is used:
		1. wrap all known symbols (visible objects) into partToChangeT[s];
			some of the structure field name might get wrapped as well, if they
			clash with visible objects,
		2. unwrap those symbols that are directly under dotT (second or later argument),
			as they are structure field names and not stand-alone objects,
		3. replace wrapped symbols into complete objectT wrappers (with types)
	*)
	
	(* 1. *)
	
	temp=expression;
	Scan[
		Function[s,
				temp=Replace[temp,s->FMTSymbols`partToChangeT[s],{-1}]
		],
		Union[visibleObjects[[All,1]]]
	];
	
	(* 2. executed for simple fields and separately for array fields,
		which are themselves wrapped in partT *)
	
	temp=FixedPoint[
		ReplaceAll[FMTSymbols`dotT[pre__,FMTSymbols`partToChangeT[s_],post___]
			->FMTSymbols`dotT[pre,s,post]
		],
		temp
	];
	temp=FixedPoint[
		ReplaceAll[FMTSymbols`dotT[pre__,FMTSymbols`partT[FMTSymbols`partToChangeT[s_],idx_],post___]
			->FMTSymbols`dotT[pre,FMTSymbols`partT[s,idx],post]
		],
		temp
	];
	
	(* 3. *)
	
	replacements=Map[
		Function[visibleObject,
			With[
				{
					objName=First[visibleObject],
					objQualifiedName=visibleObject[[2,1]],
					objType=visibleObject[[2,2]]
				},
				
				FMTSymbols`partToChangeT[objName]->FMTSymbols`objectT[objQualifiedName,objType]
			]
		],
		visibleObjects
	];
	
	temp/.replacements
]


(* Helper, transforms a single whileT condition into its
	canonical form:
	
	coreCondition && loopInvariantT[inv1, ...] && loopVariantT[var1, ...]
*)
canonicalLoopCondition[cond_]:=Module[
	{
		coreCond=True,
		loopInv=FMTSymbols`loopInvariantT[],
		loopVar=FMTSymbols`loopVariantT[]
	},
	
	If[Head[cond]===FMTSymbols`andT,
		Scan[
			Function[c, (* c is a single term from cond *)
				Switch[c,
					_FMTSymbols`loopInvariantT,
						loopInv=Join[loopInv,c],
					
					_FMTSymbols`loopVariantT,
						loopVar=Join[loopVar,c],
					
					_,
						coreCond=And[coreCond,c]
				]
			],
			cond
		],
		
		(* if the whole condition is not a conjunction, turn it into one *)
		coreCond=cond
	];
	
	FMTSymbols`andT[coreCond,loopInv,loopVar]
]

(* Helper, transforms all whileT loops into their canonical forms,
	with loop invariants and variants, for uniform processing *)
transformCanonicalWhiles[expression_]:=
	Replace[expression,FMTSymbols`whileT[cond_,body_]:>FMTSymbols`whileT[canonicalLoopCondition[cond],body],All]

(* Helper, returns the list of asserts to insert before the loop *)
invariantPreLoopAsserts[invariants_]:=Map[
	Function[inv,
		FMTSymbols`assertT[inv,"Implicit pre-loop assertion from loop invariants."]
	],
	invariants
]

(* Helper, returns the list of assumes to inject at the beginning of the loop body *)
invariantPreBodyAssumptions[invariants_]:=FMTSymbols`assumeT/@invariants

(* Helper, returns the list of asserts to inject at the end of the loop body *)
invariantPostBodyAssertions[invariants_]:=Map[
	Function[inv,
		FMTSymbols`assertT[inv,"Implicit in-loop assertion from loop invariants."]
	],
	invariants
]

(* Helper, returns the list of assumes to append after the loop *)
invariantPostLoopAssumes[cond_,invariants_]:=Map[
	Function[inv,
		FMTSymbols`assumeT[FMTSymbols`andT[FMTSymbols`notT[cond],inv]]
	],
	invariants
]

(* Helper, returns the list of ghostCaptureT directives for loop variants *)
variantGhostCaptures[ghostNames_,variants_]:=Map[
	Function[ghVar, (* ghVar is {ghostName,variant} *)
		With[
			{
				ghostName=ghVar[[1]],
				variant=ghVar[[2]]
			},
			
			FMTSymbols`ghostCaptureT[ghostName,variant]
		]
	],
	Transpose[{ghostNames,variants}]
]

(* Helper, returns the list of variant asserts *)
variantAsserts[ghostNames_,variants_]:=Map[
	Function[ghVar, (* ghVar is {ghostName,variant} *)
		With[
			{
				ghostName=ghVar[[1]],
				variant=ghVar[[2]]
			},
			
			FMTSymbols`assertT[
				variant>=0&&variant<FMTSymbols`ghostRecallT[ghostName],
				"Implicit assertion from loop variants."
			]
		]
	],
	Transpose[{ghostNames,variants}]
]

(* Helper, generates list of unique variant ghost names,
	relies on a package-level variable variantGhostIndex *)
variantGhostNames[count_]:=Table["variantGhost"<>ToString[++variantGhostIndex],{count}]

(* Helper, performs the following transformation for all loops
	in the canonical form:
	
		While[And[cond,LoopInvariant[inv],LoopVariant[var]],
			body
		]
	
	into:
	
		Assert[inv]
		While[cond,
			Assume[inv];
			body;
			Assert[inv]
		];
		Assume[And[Not[cond],inv]]
*)
transformLoopInvariants[expression_]:=Module[
	{
		e
	},
	
	(* clear the ghost counter for generating ghost names *)
	variantGhostIndex=0;
	
	e=transformCanonicalWhiles[expression];
	
	e=Replace[e,
		FMTSymbols`whileT[
			FMTSymbols`andT[
				coreCond_,
				FMTSymbols`loopInvariantT[inv___],
				FMTSymbols`loopVariantT[var___]
			],
			body_
		]
		:>
		With[
			{
				linv=List[inv],
				lvar=List[var],
				ghostNames=variantGhostNames[Length[List[var]]]
			},
			
			If[linv=={},
				(* no invariants: *)
				
				If[lvar=={},
					(* base case, no invariants and no variants present: *)
					FMTSymbols`whileT[coreCond,
						body
					],
					
					(* no invariants, but variants present: *)
					FMTSymbols`whileT[coreCond,
						FMTSymbols`compoundT[
							Sequence@@variantGhostCaptures[ghostNames,lvar],
							body,
							Sequence@@variantAsserts[ghostNames,lvar]
						]
					]
				],
				
				(* invariants present: *)
				
				If[lvar=={},
					(* invariants present, but no variants: *)
					FMTSymbols`compoundT[
						Sequence@@invariantPreLoopAsserts[linv],
						FMTSymbols`whileT[coreCond,
							FMTSymbols`compoundT[
								Sequence@@invariantPreBodyAssumptions[linv],
								body,
								Sequence@@invariantPostBodyAssertions[linv]
							]
						],
						Sequence@@invariantPostLoopAssumes[coreCond,linv]
					],
					
					(* both invariants and variants present: *)
					FMTSymbols`compoundT[
						Sequence@@invariantPreLoopAsserts[linv],
						FMTSymbols`whileT[coreCond,
							FMTSymbols`compoundT[
								Sequence@@invariantPreBodyAssumptions[linv],
								Sequence@@variantGhostCaptures[ghostNames,lvar],
								body,
								Sequence@@variantAsserts[ghostNames,lvar],
								Sequence@@invariantPostBodyAssertions[linv]
							]
						],
						Sequence@@invariantPostLoopAssumes[coreCond,linv]
					]
				]
			]
		],
		All
	];
	
	(* flatten out any nested compounds that might result from the above
		replacements; this can happen when whileT is directly enclosed
		within compoundT or when the loop body is itself compound
	*)
	e=FixedPoint[
		Function[expr,
			expr/.FMTSymbols`compoundT[pre___,FMTSymbols`compoundT[c__],post___]:>FMTSymbols`compoundT[pre,c,post]
		],
		e
	];
	
	e
]


(* Helper, replaces standard Wolfram operation symbols (Plus, Times, etc.) with local versions,
	to avoid special and implicit behaviors in pattern matching *)
sanitizeOperations[expression_]:=Module[
	{
		headReplacements,
		sanitized
	},
	
	headReplacements={
		(* user-injected comments: *)
		FMTSymbols`Comment->FMTSymbols`commentT,
		
		(* code structures: *)
		CompoundExpression->FMTSymbols`compoundT,
		If->FMTSymbols`ifT,
		Switch->FMTSymbols`switchT,
		While->FMTSymbols`whileT,
		
		(* statements: *)
		Assert->FMTSymbols`assertT,
		FMTSymbols`Assume->FMTSymbols`assumeT,
		FMTSymbols`Deduce->FMTSymbols`deduceT,
		FMTSymbols`GhostCapture->FMTSymbols`ghostCaptureT,
		FMTSymbols`GhostRecall->FMTSymbols`ghostRecallT,
		FMTSymbols`Induce->FMTSymbols`induceT,
		FMTSymbols`LoopInvariant->FMTSymbols`loopInvariantT,
		FMTSymbols`LoopVariant->FMTSymbols`loopVariantT,
		Set->FMTSymbols`assignmentT,
		
		(* expressions and their parts: *)
		Alternatives->FMTSymbols`alternativesT,
		And->FMTSymbols`andT,
		Equal->FMTSymbols`equalT,
		Greater->FMTSymbols`greaterT,
		GreaterEqual->FMTSymbols`greaterEqualT,
		Less->FMTSymbols`lessT,
		LessEqual->FMTSymbols`lessEqualT,
		Not->FMTSymbols`notT,
		Or->FMTSymbols`orT,
		Plus->FMTSymbols`plusT,
		Times[-1,v_]->FMTSymbols`negT[v], (* negation of simple terms, like in a+b-c *)
		Times[p_,Power[q_,-1]]/;q===0->FMTSymbols`divByZeroT, (* treated as grammar error *)
		Times[p_,Power[q_,-1]]/;q=!=0->FMTSymbols`divT[p,q], (* to capture simple division p/q *)
		Times->FMTSymbols`timesT,
		Unequal->FMTSymbols`unequalT,
		
		(* array indexing *)
		Part->FMTSymbols`partT,
		
		(* structure field access *)
		Dot->FMTSymbols`dotT
	};

	sanitized=FixedPoint[
		ReplaceAll[headReplacements],
		expression
	];
	
	(* stray trailing semicolons (and repeated semicolons) in compound statements are harmless,
		but create Null entries, which are just noise, except for Null statements genuinely used in
		Switch branches - for simplicity all such entries are retained.
	*)
	sanitized=sanitized/.Null->FMTSymbols`statementT[];
	
	sanitized
]


(* Helper, flattens compoundT elements, if there are any nestings *)
flattenCompounds[expression_]:=With[
	{
		flatteningReplacement=
			FMTSymbols`compoundT[pre___,between_FMTSymbols`compoundT,post___]:>
				FMTSymbols`compoundT[pre,Sequence@@between,post]
	},
	
	FixedPoint[
		ReplaceAll[flatteningReplacement],
		expression
	]
]


(* Helper, replaces calls to user-defined operations with callT wrappers,
	of the form callT[qualifiedOperationName,actualParameters] *)
sanitizeOperationCalls[expression_,visibleOperations_List]:=Module[
	{
		sanitized=expression
	},
	
	Scan[
		Function[nameRes, (* name\[Rule]qName, where name might be unqualified or qualified *)
			With[
				{
					opName=nameRes[[1]],
					qOpName=Head[nameRes[[2]]],
					formals=FMTSymbols`formalParamsT@@nameRes[[2]]
				},
				
				sanitized=sanitized/.
					opName[args___]->FMTSymbols`callT[qOpName,formals,FMTSymbols`actualParamsT[args]]
			]
		],
		visibleOperations
	];
	
	sanitized
]


(* Strips grammar wrappers and turns the given expression into a proof statement;
	for convenience, this function can handle empty expressions (Nothings) as well *)
unwrapGrammar[expression_,options_Association]:=Module[
	{
		result,
		preparationReplacements,
		mainReplacements,
		aggregateNormalizationReplacements,
		intRoundingFunction=options[FMTOptions`integerRoundingFunction]
	},
	
	preparationReplacements={
		(* simplify non-essential (internal) information: *)
		FMTSymbols`objectT[obj_,type_]->obj,
		FMTSymbols`objectParentT[obj_,type_]->obj
	};
	
	aggregateNormalizationReplacements={
		(* normalization of partT/dotT interaction is needed to ensure
			s.a[[i]] instead of (s.a)[[i]]
		*)
		FMTSymbols`partT[FMTSymbols`dotT[s_,a_],idx_]->FMTSymbols`dotT[s,FMTSymbols`partT[a,idx]]
	};
	
	mainReplacements={
		FMTSymbols`literalT[v_,_]->v,
		FMTSymbols`objectT[name_,_]->name,
		FMTSymbols`objectLhsT[name_,_]->name,
		
		(* arithmetic operations *)
		FMTSymbols`plusT->Plus,
		FMTSymbols`negT[v_]->Times[-1,v],
		FMTSymbols`timesT[r_Rational,v_]->intRoundingFunction[Times[r,v]], (* special case, as Divide[v/n] is seen as Times[Rational[1,n],v] for literal n *)
		FMTSymbols`timesT->Times,
		FMTSymbols`divT[p_,q_]->\!\(\*
TagBox[
StyleBox[
RowBox[{"IntegerPart", "[", 
RowBox[{"Times", "[", 
RowBox[{"p", ",", 
RowBox[{"Power", "[", 
RowBox[{"q", ",", 
RowBox[{"-", "1"}]}], "]"}]}], "]"}], "]"}],
ShowSpecialCharacters->False,
ShowStringCharacters->True,
NumberMarks->True],
FullForm]\),
		
		(* logic *)
		FMTSymbols`andT->And,
		FMTSymbols`notT->Not,
		FMTSymbols`orT->Or,
		
		(* relational operations *)
		FMTSymbols`equalT->Equal,
		FMTSymbols`unequalT->Unequal,
		FMTSymbols`greaterT->Greater,
		FMTSymbols`greaterEqualT->GreaterEqual,
		FMTSymbols`lessT->Less,
		FMTSymbols`lessEqualT->LessEqual,
		
		(* imperative assignments become logical statements: *)
		FMTSymbols`assignmentT->Equal,
		
		(* render array accesses in their natural syntax,
			but inactivated, otherwise Part would try to access symbols
			that are not lists in the Wolfram run-time domain: *)
		FMTSymbols`partT[array_,idx_]->Inactivate[Part[array,idx]],
		
		(* render structure field accesses in their natural syntax *)
		FMTSymbols`dotT[a_,b_]->Dot[a,b]
	};
	
	result=FixedPoint[
		ReplaceAll[preparationReplacements],
		expression
	];
	
	result=FixedPoint[
		ReplaceAll[aggregateNormalizationReplacements],
		result
	];
	
	result=FixedPoint[
		ReplaceAll[mainReplacements],
		result
	];
	
	result
]


(* Helper, identify the top-level object in the aggregate access
	by removing the inner-most access layer, which from the grammar point of view
	is closest to the root
*)
accessDropStep[object_]:=
	Switch[object,
		_Symbol,
			object,
		
		_FMTSymbols`dotT|_FMTSymbols`partT,
			object[[1]],
		
		_,
			Switch[object[[1]],
				_FMTSymbols`dotT|_FMTSymbols`partT,
					object[[1,1]],
				
				_FMTProofsFreezings`frozenT|_FMTSymbols`objectT|_FMTSymbols`objectParentT|_Symbol,
					object
			]
	]


(* Identifies the name of top-level object in the aggregate access *)
accessTopObject[object_]:=FixedPoint[accessDropStep,object]


(* Helper, returns the element type for the given name of array type *)
arrayElementType[typeHier_List,arrayType_]:=
	First[arrayType/.typeHier]


(* Helper, checks if the given type is a structure type containing the given field *)
structureFieldExistsQ[typeHier_List,qParentType_,field_Symbol]:=With[
	{
		typeHierAssoc=Association@@typeHier
	},
	
	If[Head[typeHierAssoc[qParentType]]===FMTSymbols`structure,
		With[
			{
				structureFieldAssoc=Association@@typeHierAssoc[qParentType]
			},
			
			KeyExistsQ[structureFieldAssoc,field]	
		],
		
		False
	]
]

(* Helper, returns the field type for the given structure type and field name
	(the assumption is that structureFieldExistsQ returned True) *)
structureFieldType[typeHier_List,qParentType_,field_Symbol]:=
	field/.List@@(qParentType/.typeHier)


(* Helper, transforms objectT, partT and dotT into a proper expression tree;
	these tranformations are needed, because Part binds stronger than Dot
	and s.a[[i]] is seen as Dot[s,Part[a,i]] (and then dotT[s,partT[a,i]],
	instead of a more grammar-friendly partT[dotT[s,a],i];
	this grammar-friendly form is needed for further processing (for example for array bound checks)
*)
normalizeObjects[typeHier_List,enumValues_Association,expression_]:=Module[
	{
		expr,
		replacements
	},
	
	(* dotT normalization - turn N-ary dotTs into nested binary dotTs: *)
	expr=FixedPoint[
		ReplaceAll[FMTSymbols`dotT[a_,b_,rest__]->FMTSymbols`dotT[FMTSymbols`dotT[a,b],rest]],
		expression
	];
	
	(* type transformations between objectT, dotT and partT:
	
		The following rules handle all interesting cases;
		comments use the convention:
		
		s - structure object,
		a - array object or structure field of array type,
		idx - array index (must be integer-compatible),
		f - structure field (of any type)
	*)
	replacements={
		(* a[[idx]] *)
		FMTSymbols`partT[FMTSymbols`objectT[array_,arrayType_],idx_]
			/;typesCompatibleQ[typeHier,reduceAll[typeHier,enumValues,idx],FMTSymbols`integerT]
				:>FMTSymbols`objectT[
					FMTSymbols`partT[
						FMTSymbols`objectParentT[array,arrayType],
						idx
					],
					arrayElementType[typeHier,arrayType]
				],
		
		(* s.f *)
		FMTSymbols`dotT[FMTSymbols`objectT[parent_,parentType_],field_]
			/;structureFieldExistsQ[typeHier,parentType,field]
				:>FMTSymbols`objectT[
					FMTSymbols`dotT[
						FMTSymbols`objectParentT[parent,parentType],
						field
					],
					structureFieldType[typeHier,parentType,field]
				],
		
		(* s.a[[idx]] \[Rule] element type: *)
		FMTSymbols`dotT[FMTSymbols`objectT[parent_,parentType_],FMTSymbols`partT[field_,idx_]]
			/;And[
				structureFieldExistsQ[typeHier,parentType,field],
				typesCompatibleQ[typeHier,reduceAll[typeHier,enumValues,idx],FMTSymbols`integerT]
			]
				:>FMTSymbols`objectT[
					FMTSymbols`partT[
						FMTSymbols`objectParentT[
							FMTSymbols`dotT[
								FMTSymbols`objectParentT[parent,parentType],
								field
							],
							structureFieldType[typeHier,parentType,field]
						],
						idx
					],
					arrayElementType[typeHier,structureFieldType[typeHier,parentType,field]]
				]
	};
	
	FixedPoint[
		ReplaceAll[replacements],
		expr
	]
]


(* Helper, wraps left-hand-side object into separate objectLhsT wrapper,
	to protect these objects from other non-intended replacements *)
protectLhs[expression_]:=
	expression/.FMTSymbols`assignmentT[FMTSymbols`objectT[lhs_,type_],rhs_]->FMTSymbols`assignmentT[FMTSymbols`objectLhsT[lhs,type],rhs]


(* Helper, makes the numberedT/highlightedT wrapper based on statementId and highlighted set *)
numberedAndHighlighted[statementId_,expr_,highlightedSet_]:=
	FMTSymbols`numberedT[statementId,
		If[MemberQ[highlightedSet,statementId],
			FMTSymbols`highlightedT[expr],
			expr
		]
	]


(* Helper, recursively wraps already sanitized statements into numberedT wrappers *)
processNumberedAndHighlighted[statementId_,expr_,highlightedSet_]:=Module[
	{
		e=expr,
		tempId
	},
	
	(* compound statements are not numbered as a whole,
		but their substatements are numbered individually: *)
	If[MatchQ[e,_FMTSymbols`compoundT],
		e=Map[
			Function[st,
				processNumberedAndHighlighted[statementId,st,highlightedSet]
			],
			e
		]
	];

	(* General strategy:
		1. compound statements are not numbered as a whole,
			but their substatements are numbered individually,
		2. conditional statements (ifT, switchT, whileT) are numbered in the node-first manner:
			the reason is to give focus to the conditional expression,
			which itself is the part that deserves to be numbered,
			as there can be operations that have their own verification conditions,
			branches are numbered as well, but after the main part,
		3. operation calls, assignments, assertions and assumptions are numbered
	*)
	Switch[e,
		_FMTSymbols`compoundT,
			e=Map[
				Function[st,
					processNumberedAndHighlighted[statementId,st,highlightedSet]
				],
				e
			],
		
		_FMTSymbols`ifT,
			(* remember number for the "head" of the construct
				(this is where condition is placed): *)
			tempId=++statementId;
			
			(* recurse on "then" branch: *)
			e[[2]]=processNumberedAndHighlighted[statementId,e[[2]],highlightedSet];
			
			If[Length[e]>2,
				(* recurse on "else" branch, if present *)
				e[[3]]=processNumberedAndHighlighted[statementId,e[[3]],highlightedSet]
			];
			
			(* wrap the whole: *)
			e=numberedAndHighlighted[tempId,e,highlightedSet],
		
		_FMTSymbols`switchT,
			(* remember number for the "head" of the construct
				(this is where controlling expression is placed): *)
			tempId=++statementId;
			
			(* recurse on each branch,
				note: switch branches are at odd positions, starting from 3: *)
			Scan[
				Function[branchPos,
					e[[branchPos]]=processNumberedAndHighlighted[statementId,e[[branchPos]],highlightedSet]
				],
				Range[3,Length[e],2]
			];
			
			(* wrap the whole: *)
			e=numberedAndHighlighted[tempId,e,highlightedSet],
		
		_FMTSymbols`whileT,
			(* remember number for the "head" of the construct
				(this is where condition is placed): *)
			tempId=++statementId;
			
			(* recurse on the loop body: *)
			e[[2]]=processNumberedAndHighlighted[statementId,e[[2]],highlightedSet];
			
			(* wrap the whole: *)
			e=numberedAndHighlighted[tempId,e,highlightedSet],
		
		_FMTSymbols`callT|
			_FMTSymbols`assignmentT|
			_FMTSymbols`ghostCaptureT|
			_FMTSymbols`assertT|
			_FMTSymbols`assumeT|
			_FMTSymbols`deduceT|
			_FMTSymbols`induceT,
			
			e=numberedAndHighlighted[++statementId,e,highlightedSet],
		
		_,
			(* there should be no other construct here,
				but if it happens, it can be left unprocessed (not wrapped by numberedT),
				it will not be reduced and failure will be reported anyway, elsewhere
			*)
			Null
	];
	
	e
]

SetAttributes[processNumberedAndHighlighted,HoldFirst]


resolveSyntax[expression_,preprocessor_,typeHier_List,enumValues_Association,visibleObjects_List,visibleEnums_List,visibleOperations_List,OptionsPattern[{numbered->False,highlighted->{}}]]:=Module[
	{
		e,
		highlightedSet=OptionValue[highlighted],
		synthPos,synthCode,
		statementId
	},
	
	(*
		Allow code synthesis from within the Synthesize wrapper,
		and replacement of results in the expression,
		before it is released from Hold.
	*)
	synthPos=Position[expression,_FMTSymbols`Synthesize];
	synthCode=Map[
		Function[ex, (* ex is Synthesize[userScript] *)
			ex[[1]] (* strip the Synthesize head *)
		],
		Extract[expression,synthPos]
	];
	e=ReplacePart[expression,Thread[synthPos->synthCode]];
	
	(* run the user-provided preprocessor on the given code *)
	e=preprocessor[e];
	
	(* do further grammar translations *)
	e=sanitizeOperations[e];
	e=sanitizeLiterals[e,visibleEnums];
	e=ReleaseHold[e];
	e=flattenCompounds[e];
	e=resolveObjects[e,visibleObjects];
	e=normalizeObjects[typeHier,enumValues,e];
	e=sanitizeOperationCalls[e,visibleOperations];
	e=protectLhs[e];
	
	(* transform loops containing loop invariants and variants,
		so that their conditions are checked and assumed in correct places *)
	e=transformLoopInvariants[e];
	
	If[OptionValue[numbered],
		(* replace statements with their numberedT wrappers *)
		(* Note: the following uses statementId as a shared index counter, which is updated for each replacement *)
		statementId=0;
		e=processNumberedAndHighlighted[statementId,e,highlightedSet]
	];
	
	e
]


(* Helper for reducing Switch constructs *)

switchStatementComplete[enumValues_,qEnumType_,cases_]:=Module[
	{
		oddElements,
		defaultCasePositions,
		evenElements
	},
	
	(* proper Switch statement has lockedLiteralTs at all odd positions in cases
		(perhaps with some use of alternativesT, which can be safely flattened out)
		and the default case branch is not used or is used only as the last branch,
		and statementTs[] are at all even positions *)
	oddElements=cases[[1;; ;;2]]/.FMTSymbols`alternativesT->Sequence;
	defaultCasePositions=Position[Head/@oddElements,FMTSymbols`defaultT];
	evenElements=cases[[2;; ;;2]];
	
	If[Length[oddElements]>=2&&
		ContainsOnly[Head/@oddElements,{FMTSymbols`lockedLiteralT,FMTSymbols`defaultT}]&&
		(defaultCasePositions=={}||defaultCasePositions=={{Length[oddElements]}})&&
		ContainsOnly[Last/@Cases[oddElements,_FMTSymbols`lockedLiteralT],{qEnumType}]&&
		ContainsOnly[evenElements,{FMTSymbols`statementT[]}],
		
		If[defaultCasePositions=={},
			(* all switch cases are enum literals of the proper type,
				check if all enum values are used and if they are used exactly once *)
			With[
				{
					definedEnumValues=First/@enumValues[qEnumType],
					enumValuesUsed=First/@oddElements
				},
				
				ContainsAll[enumValuesUsed,definedEnumValues]&&
					Length[enumValuesUsed]==Length[DeleteDuplicates[enumValuesUsed]]
			],
			
			(* there is one default branch
				\[Rule] other switch cases need not exhaust the list of enum values for the given enum type,
					but enum values used should not have duplicates *)
			With[
				{
					definedEnumValues=First/@enumValues[qEnumType],
					enumValuesUsed=First/@Most[oddElements]
				},
				
				ContainsOnly[enumValuesUsed,definedEnumValues]&&
					Length[enumValuesUsed]==Length[DeleteDuplicates[enumValuesUsed]]
			]			
		],
		
		False
	]
]


(* Helper functions, reduce actual parameters in procedure calls *)

(* checks whether all actual parameters were already marked as matching *)
allParamsMatchQ[formals_,actuals_]:=
	Length[formals]==Length[actuals]&&
		ContainsOnly[List@@actuals,{FMTSymbols`matchingParamT}]

(* Helper, checks matching for a single actual parameter *)
singleCallParamMatchesQ[typeHier_List,formal_,actual_]:=Module[
	{
		mode=Head[formal],
		expectedType=formal[[1,2]], (* formal is mode[name\[Rule]type] *)
		result
	},
	
	(* Note: if actual is objectT or literalT, then actual[[2]] is its type,
		otherwise, if actual is exprT, then actual[[1]] is its type. *)
	
	result=Switch[mode,
		FMTSymbols`in,
			(* formal "in" parameters can be filled
				by objects or expressions of the same or derived type;
				it is also possible to implicitly "downcast" literals to expected type
				(that is, to use literals, which are of base type, where
				the derived type is expected - verification rules will take care
				of range checking, the grammar is not constrained by this): *)

			Switch[Head[actual],
				FMTSymbols`literalT,
					(actual[[2]]===expectedType||
						actual[[2]]===(expectedType/.typeHier)), (* downcast *)
				FMTSymbols`objectT,
					(actual[[2]]===expectedType||
						(actual[[2]]/.typeHier)===expectedType), (* upcast *)
				FMTSymbols`exprT,
					(actual[[1]]===expectedType||
						(actual[[1]]/.typeHier)===expectedType), (* upcast *)
				_,
					False
			],
		
		FMTSymbols`out|FMTSymbols`inOut,
			(* formal "out" and "inOut" parameters
				can be filled with objects (of the same type) only
				Note: no downcast is allowed for simplicity,
				even though, in terms of type invariants, it would be safe for "out" params: *)
			Head[actual]===FMTSymbols`objectT&&
				actual[[2]]===expectedType
	];
	
	result
]

(* Helper, check matching for the whole signature (all parameters) *)
someCallParamsMatchQ[typeHier_List,formals_,actuals_]:=Module[
	{
		someParamsMatch=False,
		inOutActuals,
		wholeObjectAliasingDetected=False
	},
	
	If[Length[formals]==Length[actuals],
	
		(* check if individual params match their expected formal types: *)
		someParamsMatch=Or@@MapThread[
			Function[{formal,actual},
				singleCallParamMatchesQ[typeHier,formal,actual]
			],
			{List@@formals,List@@actuals}
		]
	];
	
	someParamsMatch
]

(* Helper, reduces all parameters that match at the given stage,
	assumes that the lists of formal and actual parameters have same length *)
reduceMatchingCallParams[typeHier_List,qOpName_,formals_,actuals_]:=Module[
	{
		reducedParams,
		reducedCall
	},
	
	reducedParams=FMTSymbols`actualParamsT@@MapThread[
		Function[{formal,actual},
			If[singleCallParamMatchesQ[typeHier,formal,actual],
				FMTSymbols`matchingParamT, (* reduction to match flag *)
				actual          (* no reduction in this step, leave as is *)
			]
		],
		{List@@formals,List@@actuals}
	];
	
	reducedCall=FMTSymbols`callT[qOpName,formals,reducedParams];
	
	reducedCall
]


(* Helper, returns True if all elements of exprTypes wrap compatible types
	\[Rule] types are compatible if they are same or one is derived from the other;
	Note: in this model types that are not directly related by derivation are not compatible.
*)
typesCompatibleQ[typeHier_List,exprType1_,exprType2_]:=Module[
	{
		(* strip exprT heads from both types *)
		t1=If[Head[exprType1]===FMTSymbols`exprT,exprType1[[1]],exprType1],
		t2=If[Head[exprType2]===FMTSymbols`exprT,exprType2[[1]],exprType2]
	},
	
	SameQ[t1,t2]||SameQ[t1/.typeHier,t2]||SameQ[t1,t2/.typeHier]
]

(* Helper, returns True if types can be used in a single boolean operation *)
typesBooleanCompatibleQ[exprTypes_List]:=Module[
	{
		types=Identity@@@exprTypes (* strip exprT heads from all elements *)
	},
	
	(* types are boolean-compatible if they are all booleans *)
	And@@Map[
		Function[t,
			SameQ[t,FMTSymbols`booleanT]
		],
		types
	]
]

(* Helper, returns True if types can be involved in a single arithmetic operation *)
typesArithmeticCompatibleQ[typeHier_List,exprTypes_List]:=Module[
	{
		types=Identity@@@exprTypes (* strip exprT heads from all elements *)
	},
	
	(* types are compatible it they are equivalent and based on integerT or mixed with integerT *)
	
	(* 1. check if all types except integerT are equivalent *)
	With[
		{
			allExceptInteger=Cases[types,Except[FMTSymbols`integerT],{1}]
		},
		
		SameQ@@allExceptInteger&&
			If[allExceptInteger!={},
				(First[allExceptInteger]/.typeHier)===FMTSymbols`integerT,
				True
			]
	]
]

(* Helper, returns True if types can be involved in a single relational (<,>=,etc.) operation
	(that is, they are all arithmetic) *)
typesRelationalCompatibleQ[typeHier_List,exprType1_,exprType2_]:=
	typesArithmeticCompatibleQ[typeHier,{exprType1,exprType2}]

(* Helper, returns True if types can be involved in comparison (\[Equal],\[NotEqual]) operation
	(that is, they are same or related by inheritance) *)
typesComparisonCompatibleQ[typeHier_List,exprType1_,exprType2_]:=
	typesCompatibleQ[typeHier,exprType1,exprType2]

(* Helper, returns the coerced types from the list of wrapped types *)
typeArithmeticCoerce[exprTypes_List]:=With[
	{
		types=Identity@@@exprTypes (* strip exprT heads from all elements *)
	},
	
	(* if non-integerT is used at least once in the expression, the result is non-integerT as well,
		otherwise (when all types are integerT) the result is integerT
		Note: typesArithmeticCompatibleQ, which is a condition for this transformation,
		ensures that all types other than integerT are equal *)
	With[
		{
			anyOtherThanIntegerT=Cases[types,Except[FMTSymbols`integerT],{1},1]
		},
		
		If[anyOtherThanIntegerT!={},
			First[anyOtherThanIntegerT],
			FMTSymbols`integerT
		]
	]
]


reduceStep[typeHier_List,enumValues_Association,expression_]:=Module[
	{
		result,
		replacements
	},
	
	replacements={
		(* strip numberedT wrappers: *)
		FMTSymbols`numberedT[_,st_]->st,
		
		(* alternatives (currently allowed only as alternative labels (enums) in Switch statements: *)
		FMTSymbols`alternativesT[pre___,FMTSymbols`literalT[value_,type_],post___]
			:>FMTSymbols`alternativesT[pre,FMTSymbols`lockedLiteralT[value,type],post],
		
		(* code structures (and enum locking for Switch statement);
			note: loop invariants and variants are already transformed into
				respective asserts and assumes, so do not need to be handled here *)
		FMTSymbols`compoundT[__FMTSymbols`statementT]->FMTSymbols`statementT[],
		FMTSymbols`ifT[FMTSymbols`exprT[FMTSymbols`booleanT],FMTSymbols`statementT[]]->FMTSymbols`statementT[],
		FMTSymbols`ifT[FMTSymbols`exprT[FMTSymbols`booleanT],FMTSymbols`statementT[],FMTSymbols`statementT[]]->FMTSymbols`statementT[],
		FMTSymbols`switchT[pre___,FMTSymbols`literalT[value_,type_],post___]
			:>FMTSymbols`switchT[pre,FMTSymbols`lockedLiteralT[value,type],post],
		FMTSymbols`switchT[FMTSymbols`exprT[enumType_],cases__]
			/;switchStatementComplete[enumValues,enumType,List[cases]]->FMTSymbols`statementT[],
		FMTSymbols`whileT[FMTSymbols`exprT[FMTSymbols`booleanT],FMTSymbols`statementT[]]->FMTSymbols`statementT[],
		
		(* operation calls
			Note: this transformation should be tried before objectT transformations,
				in order to capture the use of objectT or literalT as a parameter
				before it is reduced to exprT, which does not distinguish
				the original parameter format.
		*)
		FMTSymbols`callT[qOpName_,formals_,actuals_]
			/;allParamsMatchQ[formals,actuals]->FMTSymbols`statementT[],
		FMTSymbols`callT[qOpName_,formals_,actuals_]
			/;someCallParamsMatchQ[typeHier,formals,actuals]
				:>reduceMatchingCallParams[typeHier,qOpName,formals,actuals],
		
		(* assignments: *)
		FMTSymbols`assignmentT[lhs_FMTSymbols`objectLhsT,rhs_FMTSymbols`exprT]
			/;typesCompatibleQ[typeHier,lhs[[2]],rhs]
				->FMTSymbols`statementT[],
		
		(* literal is an expression of its type: *)
		FMTSymbols`literalT[_,type_]->FMTSymbols`exprT[type],
		
		(* non-lhs object is an expression, too: *)
		FMTSymbols`objectT[_,type_]->FMTSymbols`exprT[type],
		
		(* operations that are parts of expressions: *)
		
		(* negation, in principle, stays within the same type (but is subject to constraint checks) *)
		FMTSymbols`negT[n_]->n,
		
		(* arithmetic compatible types can be added and multiplied: *)
		FMTSymbols`plusT[e__FMTSymbols`exprT]/;typesArithmeticCompatibleQ[typeHier,List[e]]
			:>FMTSymbols`exprT[typeArithmeticCoerce[List[e]]],
		FMTSymbols`timesT[e__FMTSymbols`exprT]/;typesArithmeticCompatibleQ[typeHier,List[e]]
			:>FMTSymbols`exprT[typeArithmeticCoerce[List[e]]],
		FMTSymbols`divT[p_FMTSymbols`exprT,q_FMTSymbols`exprT]
			/;typesArithmeticCompatibleQ[typeHier,{p,q}]
				:>FMTSymbols`exprT[typeArithmeticCoerce[{p,q}]],
		
		(* boolean logic: *)
		FMTSymbols`andT[e__FMTSymbols`exprT]/;typesBooleanCompatibleQ[List[e]]
			:>FMTSymbols`exprT[FMTSymbols`booleanT],
		FMTSymbols`orT[e__FMTSymbols`exprT]/;typesBooleanCompatibleQ[List[e]]
			:>FMTSymbols`exprT[FMTSymbols`booleanT],
		FMTSymbols`notT[e_FMTSymbols`exprT]/;typesBooleanCompatibleQ[List[e]]
			:>FMTSymbols`exprT[FMTSymbols`booleanT],
		
		(* relational operations: *)
		FMTSymbols`equalT[p_FMTSymbols`exprT,q_FMTSymbols`exprT]
			/;typesComparisonCompatibleQ[typeHier,p,q]
				->FMTSymbols`exprT[FMTSymbols`booleanT],
		FMTSymbols`unequalT[p_FMTSymbols`exprT,q_FMTSymbols`exprT]
			/;typesComparisonCompatibleQ[typeHier,p,q]
				->FMTSymbols`exprT[FMTSymbols`booleanT],
		FMTSymbols`greaterT[p_FMTSymbols`exprT,q_FMTSymbols`exprT]
			/;typesRelationalCompatibleQ[typeHier,p,q]
				->FMTSymbols`exprT[FMTSymbols`booleanT],
		FMTSymbols`greaterEqualT[p_FMTSymbols`exprT,q_FMTSymbols`exprT]
			/;typesRelationalCompatibleQ[typeHier,p,q]
				->FMTSymbols`exprT[FMTSymbols`booleanT],
		FMTSymbols`lessT[p_FMTSymbols`exprT,q_FMTSymbols`exprT]
			/;typesRelationalCompatibleQ[typeHier,p,q]
				->FMTSymbols`exprT[FMTSymbols`booleanT],
		FMTSymbols`lessEqualT[p_FMTSymbols`exprT,q_FMTSymbols`exprT]
			/;typesRelationalCompatibleQ[typeHier,p,q]
				->FMTSymbols`exprT[FMTSymbols`booleanT],
		
		(* verification statements;
			Note: there are essentially no grammar constraints here to allow
			unrestricted use of ghost variables and ghost functions.
		*)
		
		_FMTSymbols`assertT->FMTSymbols`statementT[],
		_FMTSymbols`assumeT->FMTSymbols`statementT[],
		
		(* however, deduction and induction have strict grammar rules;
			note: induce-exists is implemented in two variants:
			- for reductions, where new, narrower statement is created from existing one,
			- for initialization of new singleton exists, from an empty-set region
			the first variant uses Not[reduction] notation and the second variant
			is valid with ForAll, too.
		*)
		FMTSymbols`deduceT[_ForAll|_Exists,instance_]->FMTSymbols`statementT[],
		FMTSymbols`induceT[_ForAll,extension_]->FMTSymbols`statementT[],
		FMTSymbols`induceT[_Exists,FMTSymbols`notT[reduction_]]->FMTSymbols`statementT[],
		FMTSymbols`induceT[(ForAll|Exists)[vars_,FMTSymbols`\[EmptySet],expr_],extension_]->FMTSymbols`statementT[],
		
		(* ghost capture also has strict rules: *)
		FMTSymbols`ghostCaptureT[name_String,_FMTSymbols`exprT]->FMTSymbols`statementT[],
		
		(* note: there is no transformation for ghostRecallT,
			it is allowed only in Asserts and Assumes *)
		
		(* user-injected comments: *)
		
		FMTSymbols`commentT[__String]->FMTSymbols`statementT[]
	};
	
	result=expression/.replacements;
	
	(* additional convenience termination: *)
	Replace[result,FMTSymbols`statementT[]->True]
]


reduceAll[typeHier_List,enumValues_Association,resolvedExpr_]:=
	FixedPoint[
		Function[expr,
			reduceStep[typeHier,enumValues,expr]
		],
		resolvedExpr
	]


grammarCorrectQ[typeHier_List,enumValues_Association,resolvedExpr_]:=
	TrueQ[reduceAll[typeHier,enumValues,resolvedExpr]]


reductionList[typeHier_List,enumValues_Association,resolvedExpr_]:=
	Most[
		FixedPointList[
			Function[expr,
				reduceStep[typeHier,enumValues,expr]
			],
			resolvedExpr
		]
	]


End[];


EndPackage[];
