(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTProofsAggregates`"];
Needs["FMTProofsBranches`"];
Needs["FMTProofsFreezings`"];
Needs["FMTSymbols`"];
Needs["FMTSystem`"];
Needs["FMTVerification`"];


BeginPackage["FMTProofs`"];


proofStatements::usage="proofStatements[sys,packageName,qOperationName,variableDomains] Returns the list of proof statements (not yet verified) for the given operation\n"<>
"and fills the domain information for all used variables.\n"<>
"Exposed in the public interface as an inspection aid.";


Begin["`Private`"];


(* Helper, iterates over all numberedT elements (at the top-level) of the given subexpression
	and accumulates path statements for them *)
iterateNumbered[
	sys_,
	packageName_,
	qOperationName_,
	opBody_,
	allPathsStatements_,
	lastCumulatedStatements_,
	branchContext_,
	cumulationWrapperFunction_,
	freezingRenames_,
	lastFreezingIds_,
	variableDomains_,
	cumulatedAggregateAccesses_,
	expr_
]:=
	Cases[expr,
		FMTSymbols`numberedT[numIndex_,numExpr_]
		:>
		processSingleNumbered[
			sys,
			packageName,
			qOperationName,
			opBody,
			allPathsStatements,
			lastCumulatedStatements,
			branchContext,
			cumulationWrapperFunction,
			freezingRenames,
			lastFreezingIds,
			variableDomains,
			cumulatedAggregateAccesses,
			numIndex/.phantomT[idx_]->idx, (* unwrap phantom, if wrapped *)
			numExpr,
			MatchQ[numIndex,_phantomT] (* is phantom *)
		],
		
		If[Head[expr]===FMTSymbols`numberedT,
			{0},
			{1}
		]
	]

SetAttributes[iterateNumbered,HoldAll]


(* Helper, reverts symbol replacements within oldT wrappers,
	freezings is a list of freezing to apply to oldT elements,
	usually empty except for operation calls, where it is equal
	to the list of freezings from the point right before the call,
	so that oldT elements (in operation post-conditions)
	can properly refer to their respective initial values *)
restoreOldValues[expression_,freezings_]:=Module[
	{
		result,
		oldTPositions=Position[expression,_FMTSymbols`oldT]
	},
	
	(* replace oldT[newSymbol] with oldT[oldSymbol],
		taking the given list of freezings into account: *)
	result=MapAt[
		Function[part,
			FMTProofsFreezings`unfreeze[part]/.freezings
		],
		expression,
		oldTPositions
	];
	
	(* finally, remove the wrappers themselves: *)
	result=result/.FMTSymbols`oldT[subExpr_]->subExpr;
	
	result
]


(* Helper, appends wrapped knowns to the list of cumulated knowns *)
appendWrappedKnowns[knowns_,lastCumulatedStatements_,cumulationWrapperFunction_]:=
	If[knowns!={},
		Scan[
			Function[known,
				AppendTo[lastCumulatedStatements,
					cumulationWrapperFunction[known]
				]
			],
			knowns
		]
	]

SetAttributes[appendWrappedKnowns,HoldRest]

(* Helper, performs recursive calls for If/While/Switch branches and generates alternative (conditional) assignments
	for those variables that have changed within those branches *)
processIfWhileSwitch[
	sys_,
	packageName_,
	qOperationName_,
	opBody_,
	allPathsStatements_,
	lastCumulatedStatements_,
	branchContext_,
	cumulationWrapperFunction_,
	freezingRenames_,
	lastFreezingIds_,
	variableDomains_,
	newVariableDomains_,
	cumulatedAggregateAccesses_,
	index_,
	expr_,
	isPhantom_
]:=Module[
	{
		head=Head[expr],
		branchesToProcess,
		conditionExpression,
		switchLabelsUsed,
		branchConditions,
		collectStrongRenamings,
		frozenKnowns,
		aggregateKnowns,
		aggregateAccessPreKnownPropagations, (* separate knowns related to array indexing and struct field accesses *)
		aggregateAccessPostKnownAssignments, (* note: there are no aggregateAccessPreKnownAssignments here, *)
		aggregateAccessPostKnownPropagations, (* because PREconditions do not generate assignments,
												and POST values include also knowns from the operation itself *)
		rememberedLastCumulatedStatements,
		branchCumulatedStatements, (* increments of lastCumulatedStatements in each branch *)
		rememberedCumulatedAggregateAccesses,
		branchCumulatedAggregateAccesses, (* increments of cumulatedAggregateAccesses in each branch *)
		rememberedFreezingRenames,
		conditionalFreezingUpdates, (* <| condition\[Rule]<| name\[Rule]freezeRenaming, ... |>, ... |> *)
		conditionalRenamingAssignments,
		condAssignments,
		localVariablesThatNeedInitialization,
		sourceObject,
		objectToRename,
		objectsBranchRenamed={}, (* set of objects that got new renamings after branch closing *)
		options=FMTSystem`fmtGetCurrentOptions[sys]
	},
	
	(* recurse on branches: *)
	
	(* Note: collect strong renamings only for If and Switch statements,
		While statements are considered to be black-boxes
		and only loop invariants can be used to export
		some knowledge from inside to outside of the loops
	*)
	Switch[head,
		FMTSymbols`ifT|FMTSymbols`ifUnrolledT,
			branchesToProcess=List@@expr[[2;;]]; (* "then" branch with "else", if present *)
			conditionExpression=FMTProofsFreezings`unwrapFrozenAndGrammar[
				FMTProofsFreezings`applyCurrentFreezings[expr[[1]],freezingRenames],
				options
			];
			
			(* consider both "then" and "else" conditions, even if the "else"
				branch does not explicitly exist - this is to enable the creation
				of "weak" variable assignments, which in turn allow values to be traced
				also outside of branches, when they are modified
			*)
			branchConditions={conditionExpression,Not[conditionExpression]};
			collectStrongRenamings=True,
		
		FMTSymbols`switchT,
			branchesToProcess=List@@expr[[3;; ;;2]]; (* 1 is enum object,
													2,4,6,... are switch labels,
													3,5,7,... are actual branches *)
			switchLabelsUsed={};
			branchConditions=Map[
				Function[caseLabel, (* caseLabel is a single enum value or a list of alternatives,
										or a blank
										in either case it is an expression that involves literalT,
										and if no literalT is present, then it was a blank *)
					With[
						{
							labels=Cases[caseLabel,_FMTSymbols`literalT,All]
						},
						
						If[labels!={},
							AppendTo[switchLabelsUsed,labels];
							Or@@Map[
								Function[singleLabel,
									FMTProofsFreezings`unwrapFrozenAndGrammar[
										FMTProofsFreezings`applyCurrentFreezings[
											FMTSymbols`equalT[expr[[1]],singleLabel],
											freezingRenames
										],
										options
									]
								],
								labels
							],
							
							(* there were no literals found in this case label,
								this must be a blank *)
							And@@Map[
								Function[singleLabel,
									FMTProofsFreezings`unwrapFrozenAndGrammar[
										FMTProofsFreezings`applyCurrentFreezings[
											FMTSymbols`unequalT[expr[[1]],singleLabel],
											freezingRenames
										],
										options
									]
								],
								Flatten[switchLabelsUsed]
							]
						]
					]
				],
				List@@expr[[2;; ;;2]]
			];
			collectStrongRenamings=True,
		
		FMTSymbols`whileT,
			branchesToProcess={expr[[2]]}; (* loop body *)
			conditionExpression=FMTProofsFreezings`unwrapFrozenAndGrammar[
				FMTProofsFreezings`applyCurrentFreezings[expr[[1]],freezingRenames],
				options
			];
			branchConditions={conditionExpression};
			collectStrongRenamings=True
	];
	
	FMTProofsBranches`beginBranches[
		rememberedLastCumulatedStatements,
		rememberedCumulatedAggregateAccesses,
		rememberedFreezingRenames,
		branchCumulatedStatements,
		branchCumulatedAggregateAccesses,
		conditionalFreezingUpdates,
		lastCumulatedStatements,
		cumulatedAggregateAccesses,
		freezingRenames
	];
	
	Scan[
		Function[branchPos,
			With[
				{
					branch=branchesToProcess[[branchPos]],
					branchCond=branchConditions[[branchPos]],
					otherBranchConditions=Delete[branchConditions,branchPos]
				},
				
				FMTProofsBranches`beginBranch[
					rememberedLastCumulatedStatements,
					rememberedCumulatedAggregateAccesses,
					rememberedFreezingRenames,
					lastCumulatedStatements,
					cumulatedAggregateAccesses,
					freezingRenames
				];
				
				(* recursive call for each identified branch: *)
				iterateNumbered[
					sys,
					packageName,
					qOperationName,
					opBody,
					allPathsStatements,
					lastCumulatedStatements,
					
					(* pre- and post-conditions have to be true
						only within the context of the branch
						and nested branches are Anded *)
					And[branchContext,branchCond],
					
					(* knowns are implied by the branch condition *)
					Function[st,
						cumulationWrapperFunction[
							Implies[branchCond,st]
						]
					],
					
					freezingRenames,
					lastFreezingIds,
					variableDomains,
					cumulatedAggregateAccesses,
					branch
				];
				
				FMTProofsBranches`endBranch[
					branchCond,
					otherBranchConditions,
					rememberedLastCumulatedStatements,
					rememberedCumulatedAggregateAccesses,
					rememberedFreezingRenames,
					branchCumulatedStatements,
					branchCumulatedAggregateAccesses,
					conditionalFreezingUpdates,
					lastCumulatedStatements,
					cumulatedAggregateAccesses,
					freezingRenames,
					collectStrongRenamings,
					True (* always collect weak renamings (those that bypass the branch) *)
				]
			]
		],
		
		Range[Length[branchesToProcess]]
	];
	
	(* conditionalFreezingUpdates has the form:
		
		<| condition \[Rule] <|name\[Rule]freezeRenaming, ...|>, ... |>
		
		The meaning of this structure is that when condition is True,
		the symbol name got new replacement; all branches need to be taken into account.
		The renamings, wrapped in implications based on branch conditions,
		are added to allPathsStatements, so that
		further statements can rely on this in their own proofs.
		The following call to endBranches does this.
	*)
	
	conditionalRenamingAssignments=FMTProofsBranches`endBranches[
		rememberedLastCumulatedStatements,
		rememberedCumulatedAggregateAccesses,
		rememberedFreezingRenames,
		branchCumulatedStatements,
		branchCumulatedAggregateAccesses,
		conditionalFreezingUpdates,
		lastCumulatedStatements,
		cumulatedAggregateAccesses,
		freezingRenames,
		isPhantom
	];
	
	(* When isPhantom is True, the branch is joined in the artificial context,
		from which statements are not recorded, and from which conditional
		renamings should not be performed either. That is, assignments from the joined branch
		should become visible in the context after the branch without any additional
		renamings. This is used for unrolled While loops, which are unrolled into
		a contruct that has additional If statement embracing the original While
		and where there is no nee for two separate renamings.
		The visibility of names was already taken care of by endBranches above,
		which, within the phantom context, does not revert the freezingRenames structure
		and keeps the one cumulated within the branch.
		Note: this works when there was only one joining branch
			(which is the case for unrolled loop).
	*)
	If[Not[isPhantom],
		(* process additional assignments that result from alternative renamings in branches: *)
		
		localVariablesThatNeedInitialization=
			First/@FMTSystem`operationUninitializedLocals[sys,qOperationName];
		Scan[
			Function[cond,
				condAssignments=conditionalRenamingAssignments[cond];
				Scan[
					Function[condAssignment,
						objectToRename=condAssignment[[1]];
						sourceObject=condAssignment[[2]];
						
						(* processing steps are as above for regular expressions: *)
						
						frozenKnowns=FMTProofsFreezings`freezeAndReplaceSingleAssignment[
							condAssignment,
							freezingRenames,
							lastFreezingIds,
							variableDomains,
							newVariableDomains,
							options,
							
							(* create new freeze for lhs if it was not yet done,
								but reuse the already created freeze for later assignments
								involving the same lhs object: *)
							makeNewLhsFreeze->FreeQ[objectsBranchRenamed,objectToRename],
							
							(* rhs in this assignment already properly
								indicates the right frozen source value,
								so there is no need to apply the freezings again: *)
							applyRhsFreezings->False
						];
						
						AppendTo[objectsBranchRenamed,objectToRename];
						
						aggregateKnowns=FMTProofsAggregates`processAggregateAccesses[
							sys,
							lastCumulatedStatements,
							frozenKnowns,
							cumulatedAggregateAccesses
						];
						aggregateAccessPostKnownAssignments=aggregateKnowns[[1]];
						aggregateAccessPostKnownPropagations=aggregateKnowns[[2]];
						
						aggregateAccessPostKnownAssignments=
							FMTProofsFreezings`unwrapFrozenAndGrammar[
								aggregateAccessPostKnownAssignments,
								options
							];
						aggregateAccessPostKnownPropagations=
							FMTProofsFreezings`unwrapFrozenAndGrammar[
								aggregateAccessPostKnownPropagations,
								options
							];
						
						aggregateAccessPostKnownAssignments=
							DeleteCases[aggregateAccessPostKnownAssignments,True];
						
						aggregateAccessPostKnownPropagations=
							DeleteCases[aggregateAccessPostKnownPropagations,True];
						
						appendWrappedKnowns[
							aggregateAccessPostKnownAssignments,
							lastCumulatedStatements,
							Function[st,
								cumulationWrapperFunction[Implies[cond,st]]
							]
						];
						
						(* add also the conditional propagation of initialization flag,
							but only for those local variables that are initially uninitialized: *)
						If[MemberQ[localVariablesThatNeedInitialization,objectToRename[[1]]],
							AppendTo[
								lastCumulatedStatements,
								cumulationWrapperFunction[
									Implies[
										cond,
										Equal@@
											FMTSymbols`initializedT/@
												FMTProofsFreezings`unwrapFrozenAndGrammar[frozenKnowns[[1]],options]
									]
								]
							]
						]
					],
					condAssignments
				]
			],
			Keys[conditionalRenamingAssignments]
		]
	]
]

SetAttributes[processIfWhileSwitch,HoldAll]

(* Helper, adds new domain information to the main set;
	performs grammar unwrapping and domain replacements as well
*)
appendToDomains[sys_,variableDomains_,newVariableDomains_,options_]:=Module[
	{
		typeHier=FMTSystem`typeHierarchy[sys]
	},
	
	AppendTo[variableDomains,
		Map[
			Function[vd, (* vd is variable\[Rule]domain *)
				With[
					{
						varName=vd[[1]],
						domain=vd[[2]]
					},
					
					FMTProofsFreezings`unwrapFrozenAndGrammar[varName,options]->
						(domain/.typeHier)/.
							{FMTSymbols`booleanT->Booleans,FMTSymbols`integerT->Integers}
				]
			],
			Normal[newVariableDomains]
		]
	];
]

SetAttributes[appendToDomains,HoldAll]

(* Helper, flattens knowns that are bound by conjunctions and removes duplicates.
	Note: it is convenient for knowns to be broken down as a list instead of being
		compound within the andT statements; that is, knownsForThisPart is a list of knowns,
		and those knowns that happen to be andTs are broken and flattened out, so that for example
		{a,b&&c,d&&e} becomes {a,b,c,d,e}
	Note: at this stage knowns are already grammar-unwrapped,
		so the flattening is on And instead of andT.
*)
flattenKnowns[knowns_]:=DeleteDuplicates[
	Flatten[
		Replace[knowns,And->List,{2},Heads->True],
		1
	]
]	

(* Helper, simplifies the list of knowns by removing duplicates
	and resolving basic implications;
	Note: for readability, it makes sense to retain the order of knowns,
	so it is not possible to just run Simplify on the whole list,
	as this operation has too much impact on the expression format;
	instead, knowns are simplified one-by-one and in the case of implications,
	the left-hand-side is checked with all remaining knowns to see whether
	it reduces to True, in which case the implication itself is reduced
	to its right-hand-side only - this is repeated if possible *)
simplifyKnowns[knowns_]:=Module[
	{
		knownsFlattened=flattenKnowns[knowns],
		i,
		implExpr, (* consecutive steps in simplifying implications *)
		implicationFullySimplified,
		singleResult
	},
	
	i=0;
	Map[
		Function[k, (* k is a single known *)
			i=i+1;
			
			If[Head[k]===Implies,
				(* simplify implications based on other entries *)
				
				implExpr=k;
				implicationFullySimplified=False;
				While[Not[implicationFullySimplified],
					With[
						{
							(* k is lhs=>rhs *)
							lhs=implExpr[[1]],
							rhs=implExpr[[2]],
							remainingKnowns=Delete[knownsFlattened,i]
						},
						
						(* try to simplify left-hand-side: *)
						singleResult=Simplify[Implies[And@@remainingKnowns,lhs]];
						If[singleResult===True,
							(* lhs is a consequence of the other knowns,
								reduce the implication to rhs only *)
							implExpr=rhs;
							
							(* is it still an implication (it might have been nested)?
								if not, finish *)
							If[Head[implExpr]=!=Implies,
								singleResult=Simplify[implExpr];
								implicationFullySimplified=True
							],
							
							(* lhs is not a consequence of other knowns,
								but it might contradict it: *)
							singleResult=Simplify[Implies[And@@remainingKnowns,Not[lhs]]];
							If[singleResult===True,
								(* lhs contradicts the other knowns,
									reduce the whole implication to True *)
								singleResult=True;
								implicationFullySimplified=True,
							
								(* lhs cannot be simplified *)
								singleResult=implExpr;
								implicationFullySimplified=True
							]
						]
					]
				],
				
				singleResult=Simplify[k];
			];
				
			If[singleResult===True,
				Nothing,
				singleResult
			]
		],
		knownsFlattened
	]
]

(* Helper, stores captured ghost expressions,
	gcs is {ghostCaptureT[name,expr],...} *)
storeCapturedGhosts[gcs_]:=AppendTo[capturedGhosts,Rule@@@gcs]

(* Helper for recursive calls, processes a single numberedT expression:
	get fact and wish statements related to that part
	and accumulates them on the path from the beginning of the operation's body
	(the statements are pre-processed by the cumulation wrapper function);
	accumulates all such statement lists with index numbers as well
*)
processSingleNumbered[
	sys_,
	packageName_,
	qOperationName_,
	opBody_,
	allPathsStatements_,
	lastCumulatedStatements_,
	branchContext_,
	cumulationWrapperFunction_,
	freezingRenames_,
	lastFreezingIds_,
	variableDomains_,
	cumulatedAggregateAccesses_,
	index_,
	expr_,
	isPhantom_
]:=Module[
	{
		statementsForThisPart, (* both wishes and knowns for the given points *)
		preCondsForThisPart, (* preconditions *)
		postCondsForThisPart, (* postconditions *)
		preCallFreezings={}, (* freezing from before the operation call (used only with calls) *)
		knownsForThisPart, (* new statements resulting from object assignments *)
		frozenKnowns={}, (* additional facts related to freezing of assigned variables,
							these should be further processes by aggregate access analysis *)
		typesOfNewFreezes={},

		aggregateKnowns, (* temp for aggregate analysis *)
		aggregateAccessPreKnownPropagations, (* separate knowns related to array indexing and struct field accesses *)
		aggregateAccessPostKnownAssignments, (* note: there are no aggregateAccessPreKnownAssignments here, *)
		aggregateAccessPostKnownPropagations, (* because PREconditions do not generate assignments,
												and POST values include also knowns from the operation itself *)
		
		newVariableDomains=<||>, (* domains of new variables that appear in knowns due to freezing *)
		
		canOptimizeBranchCondition, (* whether new known refers to new symbols only
										and does not need wrapping into the branch condition *)
		options=FMTSystem`fmtGetCurrentOptions[sys]
	},
	
	(* get verification statements for this index: *)
	statementsForThisPart=FMTVerification`operationConditions[sys,packageName,qOperationName,opBody,index];

	(* statements are {preconds,postconds,knowns}: *)
	preCondsForThisPart=statementsForThisPart[[1]];
	postCondsForThisPart=statementsForThisPart[[2]];
	knownsForThisPart=statementsForThisPart[[3]];
	
	(* pre-conditions are resolved using the old freezing renames
		(before assigned variables get new symbol name): *)
	preCondsForThisPart=FMTProofsFreezings`applyCurrentFreezings[preCondsForThisPart,freezingRenames];
	
	(* freeze the symbol that gets new value in assignment: *)
	If[MatchQ[expr,_FMTSymbols`assignmentT],
		frozenKnowns=FMTProofsFreezings`freezeAndReplaceSingleAssignment[
			expr,
			freezingRenames,
			lastFreezingIds,
			variableDomains,
			newVariableDomains,
			options
		]
	];
	
	(* create new freezings for symbols that are used
		as inOut or out actual parameters in operation calls: *)
	If[MatchQ[expr,_FMTSymbols`callT],
		preCallFreezings=freezingRenames;
		
		(* expr is callT[qOpName,formalParamsT[...],actualParamsT[...]] *)
		(* here ScanThread was intended, but it does not exist: *)
		MapThread[
			Function[{formal,actual},
				With[
					{
						mode=Head[formal]
					},
					
					If[mode===FMTSymbols`inOut||mode===FMTSymbols`out,
						(* Note: no new knowns are created here, as was the case above
							with assignments; the reason for this is that assignments are
							direct sources of knowns (values for newly frozen symbols),
							whereas in operation calls the only knowns about the output
							parameters come from operation's post-conditions
							and from formal parameter type, which
							are both dealt with elsewhere (in FMTVerification`) *)
						
						FMTProofsFreezings`freezeSingleSymbol[
							actual,freezingRenames,lastFreezingIds,
							variableDomains,newVariableDomains,options]
					]
				]
			],
			{List@@expr[[2]],List@@expr[[3]]}
		]
	];
	
	(* post-conditions and knowns should refer to new symbol name,
		if it was frozen in assignment: *)
	postCondsForThisPart=FMTProofsFreezings`applyCurrentFreezings[postCondsForThisPart,freezingRenames];
	knownsForThisPart=FMTProofsFreezings`applyCurrentFreezings[knownsForThisPart,freezingRenames];
	
	(* at this point, when variables are already properly versioned,
		ghost captures can be processed and removed from the set of post-conditions: *)
	storeCapturedGhosts[Cases[postCondsForThisPart,gc_FMTSymbols`ghostCaptureT]];
	postCondsForThisPart=DeleteCases[postCondsForThisPart,gc_FMTSymbols`ghostCaptureT];
	
	(* similarly, ghost references (GhostRecall) can be processed here as well
		in all three sets (pre-, post-, knowns): *)
	preCondsForThisPart=preCondsForThisPart/.FMTSymbols`ghostRecallT[name_String]:>capturedGhosts[name];
	postCondsForThisPart=postCondsForThisPart/.FMTSymbols`ghostRecallT[name_String]:>capturedGhosts[name];
	knownsForThisPart=knownsForThisPart/.FMTSymbols`ghostRecallT[name_String]:>capturedGhosts[name];
	
	(* if knowns are {initializedT[frozenT[...]]}, then it is an initialization
		of new variable freeze and this does not need to be wrapped in cumulationWrapperFunction
		when it is added to the list of knowns
	*)
	If[MatchQ[
		knownsForThisPart,
		List[FMTSymbols`initializedT[_FMTProofsFreezings`frozenT]]],
		
		canOptimizeBranchCondition=True;
		typesOfNewFreezes=Cases[
			knownsForThisPart,
			f_FMTProofsFreezings`frozenT:>{Subscript[f[[1,1]],f[[2]]],f[[1,2]]},
			All
		],
		
		canOptimizeBranchCondition=False
	];
	
	(* separate handling of aggregate accesses,
		operate only on those accesses that were used in collected proof statements;
		Note: the knowns related to aggregate accesses are divided into two groups:
			- known assignments, which result from aggregate assignments (these knowns
				reflect the propagation of values for individual fields and array elements),
				these are stored in lastCumulatedStatements and can be used in proofs
				generated immediately or for later points in the operation body,
			- known propagations, which are relations between fields and elements within
				the same aggregate object and involve array indexing,
				these are not stored permanently and are recreated every time,
				as their set will depend on the cumulated set of access paths that
				were used within the given operation
				(that is, instead of storing these knowns in lastCumulatedStatements,
				the cumulatedAggregateAccesses is a source to recreate them).
	*)
	If[Not[isPhantom],
		With[
			{
				simplifiedCumulatedStatements=simplifyKnowns@Join[
					lastCumulatedStatements,
					If[branchContext=!=True,List[branchContext],{}]
				]
			},
			
			aggregateKnowns=FMTProofsAggregates`processAggregateAccesses[
				sys,
				simplifiedCumulatedStatements,
				preCondsForThisPart[[All,2]],
				cumulatedAggregateAccesses
			];
			
			(* note: preconditions do not generate new assignments,
				only aggregate accesses that might require new value propagation knowns,
				-> there are no aggregateAccessPreKnownAssignments,
				and aggregateKnowns[[1]] is supposed to be empty *)
			Assert[aggregateKnowns[[1]]=={}];
			aggregateAccessPreKnownPropagations=aggregateKnowns[[2]];
			
			aggregateKnowns=FMTProofsAggregates`processAggregateAccesses[
				sys,
				simplifiedCumulatedStatements,
				Join[
					frozenKnowns,
					knownsForThisPart,
					postCondsForThisPart[[All,2]]
				],
				cumulatedAggregateAccesses
			];
			
			aggregateAccessPostKnownAssignments=aggregateKnowns[[1]];
			aggregateAccessPostKnownPropagations=aggregateKnowns[[2]];
			
			(* aggregateAccessPostKnownPropagations might contain new entries
				that need to have their domain identified *)
			FMTProofsAggregates`identifyNewVariableDomains[
				aggregateAccessPostKnownPropagations,
				newVariableDomains
			]
		]
	];
	
	(* special handling of oldT[expr] - revert symbol replacements within this wrapper,
		note: this has to be done before unwrapping grammar, as this process relies
		on all wrappers to be in place *)
	preCondsForThisPart=restoreOldValues[preCondsForThisPart,preCallFreezings];
	knownsForThisPart=restoreOldValues[knownsForThisPart,preCallFreezings];
	postCondsForThisPart=restoreOldValues[postCondsForThisPart,preCallFreezings];
	
	(* all knowns and statements are grammar-wrapped and these have to be unwrapped
		before passing them to the prover: *)
	preCondsForThisPart=FMTProofsFreezings`unwrapFrozenAndGrammar[preCondsForThisPart,options];
	postCondsForThisPart=FMTProofsFreezings`unwrapFrozenAndGrammar[postCondsForThisPart,options];
	knownsForThisPart=FMTProofsFreezings`unwrapFrozenAndGrammar[knownsForThisPart,options];
	aggregateAccessPreKnownPropagations=FMTProofsFreezings`unwrapFrozenAndGrammar[aggregateAccessPreKnownPropagations,options];
	aggregateAccessPostKnownAssignments=FMTProofsFreezings`unwrapFrozenAndGrammar[aggregateAccessPostKnownAssignments,options];
	aggregateAccessPostKnownPropagations=FMTProofsFreezings`unwrapFrozenAndGrammar[aggregateAccessPostKnownPropagations,options];
	
	(* sanitize preconditions and knowns by removing tautologies and duplicates: *)
	preCondsForThisPart=DeleteDuplicates[
		DeleteCases[preCondsForThisPart,
			{_String,FMTSymbols`alwaysT[True]}|{_String,FMTSymbols`neverT[False]}
		]
	];
	aggregateAccessPreKnownPropagations=DeleteCases[aggregateAccessPreKnownPropagations,True];
	aggregateAccessPostKnownAssignments=DeleteCases[aggregateAccessPostKnownAssignments,True];
	aggregateAccessPostKnownPropagations=DeleteCases[aggregateAccessPostKnownPropagations,True];
	
	(* Note: preconditions are considered *without* new knowns that are contributed
		at this point - the reason for this is that new knowns might divert the solver
		from finding examples (or counter-examples) for the precondition,
		as they can unnecessarily constrain the solution domain, for example:
		
		y=1/x;
		
		The precondition for the above is x\[NotEqual]0 and the new known
		(with frozen symbol for y) is y$n=1/x. If the domain for both {x,y} is Integers,
		then the counterexample for x\[Equal]0 will not be found, as the new known implicitly
		constrains the solution domain to legal values of x.
		
		The solution to this problem is to gather preconditions without new knowns
		and then, separately, postconditions with new knowns. This also means
		that in the complete result there can be 1 or 2 entries for the given index.
	*)
	
	(* 1. preconditions without new knowns,
		but with current branch context, if non-trivial: *)
	
	If[Not[isPhantom],
		If[preCondsForThisPart!={},
			AppendTo[allPathsStatements,
				{
					index,
					simplifyKnowns@Join[
						lastCumulatedStatements,
						aggregateAccessPreKnownPropagations,
						If[branchContext=!=True,List[branchContext],{}]
					],
					preCondsForThisPart
				}
			]
		]
	];
	
	(* 2. accumulate new knowns: *)
	
	(* Note: knownsForThisPart are accumulated after wrapping in cumulationWrapperFunction,
		because these are new facts about potentially reused symbols,
		and the facts should respect the current conditional branch [* but see below];
		the aggregateXXX facts are accumulated without any wrapping (Identity),
		because they refer to new versions of some symbols and this way do not
		conflict with any existing knowns - wrapping them in cumulationWrapperFunction
		would not be harmful, but would unnecessarily clutter the presentation with
		unnecessary conditional implications.
		
		[*] the only exception being initializedT[frozenT[...]], which refers to the
		newly created symbol and as such does not need wrapping in the current branch condition,
		this opportunity is detected earlier and stored in canOptimizeBranchCondition
	*)
	If[Not[isPhantom],
		appendWrappedKnowns[knownsForThisPart,lastCumulatedStatements,
			If[canOptimizeBranchCondition,
				(* optimization for initializedT[frozenT[...]], no need for wrapping *)
				Identity,
				
				(* wrap any other expression in the current branch condition *)
				cumulationWrapperFunction
			]
		];
		appendWrappedKnowns[aggregateAccessPostKnownAssignments,lastCumulatedStatements,Identity];
		appendWrappedKnowns[aggregateAccessPostKnownPropagations,lastCumulatedStatements,Identity],
		
		(* even for phantom expressions, initialization flags should be retained,
			so that further verification statements can refer to them when needed;
			in addition, type invariants should be re-defined for these symbols,
			because without any history there no knowledge whatsoever with regard
			to their set of possible values;
			type invariants play the role of the most conservative and minimal guess *)
		appendWrappedKnowns[
			Join[
				Cases[knownsForThisPart,_FMTSymbols`initializedT,All],
				Map[
					Function[vt, (* vt is {Subscript[varName,version], qType} *)
						With[
							{
								var=vt[[1]],
								qType=vt[[2]],
								invariants=FMTSystem`typeInvariants[sys,vt[[2]]]
							},
							
							invariants/.qType->var
						]
					],
					typesOfNewFreezes
				]
			],
			lastCumulatedStatements,
			Identity
		]
	];
	
	(* 3. postconditions take into account new knowns
		and the current branch context, if non-trivial: *)
	
	If[Not[isPhantom],
		If[(postCondsForThisPart!={})||
			(knownsForThisPart!={})||
			(aggregateAccessPostKnownAssignments!={})||
			(aggregateAccessPostKnownPropagations!={}),
			
			AppendTo[allPathsStatements,
				{
					index,
					simplifyKnowns@Join[
						lastCumulatedStatements,
						If[branchContext=!=True,List[branchContext],{}]
					],
					postCondsForThisPart
				}
			]
		]
	];
	
	(* If the part is a nested numberedT (for example, If, While, Switch, etc.),
		then remember current value of lastCumulatedStatements
		and recurse on all available branches;
		revert to the remembered value after recursion
	*)
	If[MemberQ[
		{FMTSymbols`ifT,FMTSymbols`ifUnrolledT,FMTSymbols`whileT,FMTSymbols`switchT},
		Head[expr]],
		
		processIfWhileSwitch[
			sys,
			packageName,
			qOperationName,
			opBody,
			allPathsStatements,
			lastCumulatedStatements,
			branchContext,
			cumulationWrapperFunction,
			freezingRenames,
			lastFreezingIds,
			variableDomains,
			newVariableDomains,
			cumulatedAggregateAccesses,
			index,
			expr,
			isPhantom
		]
	];
	
	(* add, with proper conditioning (unwrap grammar, replace hierarchy, etc.),
		domains of all new variables that appeared in knowns due to freezing *)
	appendToDomains[sys,variableDomains,newVariableDomains,options];
]

SetAttributes[processSingleNumbered,HoldAll]


(* Helper, prepares a list of paths with logic statements from the beginning of the operation
	to the index where the path ends
*)
iteratePathsAndGetFactStatements[
	sys_,
	packageName_,
	qOperationName_,
	opBody_,
	variableDomains_
]:=Module[
	{
		freezingRenames=<||>,
		lastFreezingIds=<||>,
		lastCumulatedStatements={}, (* {statement,...} *)
		cumulatedAggregateAccesses=<||>,
		allPathsStatements={} (* {{index,{statement,...}},...} *)
	},
	
	(* Note: allPathsStatements is updated by calls to processSingleNumbered *)
	
	(* 1. gather statements for the local "start" point
		(for example: variables are not initialized at the start): *)
	processSingleNumbered[
		sys,
		packageName,
		qOperationName,
		opBody,
		allPathsStatements,
		lastCumulatedStatements,
		True, (* dummy branch context at the operation's top-level *)
		Identity, (* dummy cumulation wrapper at the operation's top-level *)
		freezingRenames,
		lastFreezingIds,
		variableDomains,
		cumulatedAggregateAccesses,
		"start",
		Null, (* dummy *)
		False
	];
	
	(* 2. process recursively all numberedT in the operation's body: *)
	iterateNumbered[
		sys,
		packageName,
		qOperationName,
		opBody,
		allPathsStatements,
		lastCumulatedStatements,
		True, (* dummy branch context at the operation's top-level *)
		Identity, (* dummy cumulation wrapper at the operation's top-level *)
		freezingRenames,
		lastFreezingIds,
		variableDomains,
		cumulatedAggregateAccesses,
		opBody (* start with the whole operation's body (iterations are done at level {1}),
					further recursions (on conditional branches and loops) will narrow this scope *)
	];
	
	(* 3. finally, accumulate operation post-conditions: *)
	processSingleNumbered[
		sys,
		packageName,
		qOperationName,
		opBody,
		allPathsStatements,
		lastCumulatedStatements,
		True, (* dummy branch context at the operation's top-level *)
		Identity, (* dummy cumulation wrapper at the operation's top-level *)
		freezingRenames,
		lastFreezingIds,
		variableDomains,
		cumulatedAggregateAccesses,
		"finish",
		Null, (* dummy *)
		False
	];
	
	allPathsStatements
]

SetAttributes[iteratePathsAndGetFactStatements,HoldAll]


(* singleUnrollLoop and preprocessLoopConditions are helpers
	that transform whileT loops into:
	
	ifUnrolledT[cond,
		phantomLoopBody,
		whileT[cond,
			body
		]
	]
	
	Statements from phantomLoopBody are not added to allPathStatements,
	but they do generate new frozen names.
	The index numbers are preserved from the original loop to ensure
	there is no impact on the flow diagram.
	Note: the whileT[cond,body] as a whole is also phantom, to avoid
	duplicated pre-conditions, which are already taken into account by ifUnrolledT[],
	the body inside the loop is not phantom.
*)
singleUnrollLoop[index_,cond_,body_]:=Module[
	{
		phantomBody=body
			/.FMTSymbols`numberedT[numIdx_,numExpr_]->
				FMTSymbols`numberedT[phantomT[numIdx],numExpr],
		
		furtherLoopIterations=FMTSymbols`numberedT[phantomT[index],
			FMTSymbols`whileT[cond,body] (* loop construct repeated *)
		]
	},
	
	FMTSymbols`numberedT[index,
		FMTSymbols`ifUnrolledT[cond, (* original condition *)
			
			If[Head[phantomBody]===FMTSymbols`compoundT,
				Append[
					phantomBody,
					furtherLoopIterations
				],
				
				FMTSymbols`compoundT[
					phantomBody,
					furtherLoopIterations
				]
			]
		]
	]
]

preprocessLoopConditions[expr_]:=expr/.
	FMTSymbols`numberedT[index_,FMTSymbols`whileT[cond_,body_]]:>
		singleUnrollLoop[index,cond,body]


(* Returns the list of proof statements in the form:
	{
		{
			index,
			knowns,
			{{description,wish},...}
		},
		...
	},
	
	where index identifies the place in the operation where the proof statement was generated
	(this can be used to highlight that place in grammar or control flow diagrams),
	where knowns is a list of frozen proof statements until (and excluding) the "wish" point,
	and wishes is a list of {description,wish} pairs,
	where wish is a statement that needs to be true in order for the operation to be correct.
	
	In the simplest case, the whole list might be empty - for example when the operation itself
	has no postconditions (that is, nothing to wish for at the end of operation) and no precondition
	in any of its internal statements.
	If the operation has postconditions, these are wishes that accompany all knowns from the whole
	operation.
	Note: wishes have symbols replaced with freezingRenames that were active at the wish point.
	
	In addition, variableDomains is an Association that is filled with domain names
	for all variables used in the proof statements.
*)
proofStatements[sys_,packageName_,qOperationName_,variableDomains_]:=Module[
	{
		opBody, (* resolved operation body *)
		typeHier,
		domains, (* temporary, to find variable domains as used in the operation body and proof statements *)
		allPathStatements={}, (* {{index,facts,{{description,wish},...},reverseFreezes},...} - list of known facts
								from the beginning of the operation to the given point,
								together with their verification conditions (wishes);
								indices include "finish" *)
		
		options=FMTSystem`fmtGetCurrentOptions[sys]
	},
	
	opBody=FMTSystem`resolvedOperationBody[sys,packageName,qOperationName,True,{}];
	
	opBody=preprocessLoopConditions[opBody];
	
	(* opBody can be Null if the given operation is declared, but not defined *)
	If[opBody=!=Null,
		typeHier=FMTSystem`typeHierarchy[sys];
		
		(* extract domains for those objects that are used in the operation: *)
		domains=FMTSystem`operationAllMentionedObjects[sys,packageName,qOperationName,opBody];
		
		(* turn program-level types into proof domain names: *)
		appendToDomains[sys,variableDomains,domains,options];
		
		(* clear the list of ghost exporessions *)
		capturedGhosts=<||>;
		
		allPathStatements=iteratePathsAndGetFactStatements[
			sys,
			packageName,
			qOperationName,
			opBody,
			variableDomains
		]
	];
	
	allPathStatements
]

SetAttributes[proofStatements,HoldAll]


End[];


EndPackage[];
