(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTGrammar`"];
Needs["FMTSymbols`"];


BeginPackage["FMTProofsFreezings`"];


unfreeze::usage="unfreeze[expression] Strips all frozenT wrappers.";


unwrapFrozen::usage="unwrapFrozen[expression] Changes internal frozenT wrapper into huma-readable Subscript.";


unwrapFrozenAndGrammar::usage="unwrapFrozenAndGrammar[expression,options] Unwraps both frozenT and all internal grammar wrappers to create human-readable representation.";


applySingleFreezing::usage="applySingleFreezing[expression,name,freezingId] Replaces name with its frozen version within expression.";


applyCurrentFreezings::usage="applyCurrentFreezings[expression,freezingRenames] Applies all freezing replacements within the given expression.";


freezeSingleSymbol::usage="freezeSingleSymbol[name,freezingRenames,lastFreezingIds,variableDomains,newVariableDomains,options] Creates new freezing for the given symbol\n"<>
"and maintains freezing accounting structures.";


freezeAndReplaceSingleAssignment::usage="freezeAndReplaceSingleAssignment[a,freezingRenames,lastFreezingIds,variableDomains,newVariableDomains,options] Freezes and replaces symbols in a single assignment operation\n"<>
"and maintains freezing accounting structures.";


(* these definitions are placed here to force all used symbols in the public context of this package: *)

proofsFreezingsTags={
	frozenT (* wrapper for frozen symbols and expressions, rendered as subscript *)
};


Begin["`Private`"];


(* Removes the freezing wrappers *)
unfreeze[expression_]:=expression/.frozenT[name_,_]->name


(* Translates frozenT to Subscript,
	performs only small fraction of the whole unwrapFrozenAndGrammar functionality,
	which deal with the whole grammar *)
unwrapFrozen[expression_]:=With[
	{
		replacements={
			(* render frozen symbols as subscripted names: *)
			frozenT[name_,0]->name,
			frozenT[name_,id_]->Subscript[name,id]
		}
	},
	
	expression/.replacements
]


(* Strips grammar wrappers and turns the given expression into a proof statement;
	for convenience, this function can handle empty expressions (Nothings) as well *)
unwrapFrozenAndGrammar[expression_,options_Association]:=Module[
	{
		result,
		unfreezeReplacements
	},
	
	unfreezeReplacements={
		(* render frozen symbols as subscripted names: *)
		frozenT[name_,0]->name,
		frozenT[name_,id_]->Subscript[name,id]
	};
	
	result=FMTGrammar`unwrapGrammar[expression,options];
	
	result=FixedPoint[
		ReplaceAll[unfreezeReplacements],
		result
	];
	
	result
]


(* Creates a new name for a symbol that will be frozen
	and updates the freezingRenames structure appropriately. *)
freezingReplacement[name_,lastIds_,freezingRenames_]:=Module[
	{
		lookupName, (* name used to look up existing replacements might be a simplified form
						with regard to the given full expression name
						- in particular, arrays get new freezes regardless
						of their indexing component *)
		id,
		origName, (* simplified name, but in its original wrapper (object, parent or lhs) *)
		freezingRule,
		objName,freezingRuleForObj,
		parentName,freezingRuleForParent,
		lhsName,freezingRuleForLhs
	},
	
	origName=FMTGrammar`accessTopObject[name];
	
	lookupName=FMTSymbols`objectT@@origName;
	
	id=Lookup[lastIds,lookupName,0]+1;
	lastIds[lookupName]=id;
	
	(* create freezing rules for all cases: objectT, objectParentT, objectLhsT: *)
	
	objName=FMTSymbols`objectT@@lookupName;
	freezingRuleForObj=objName->frozenT[objName,id];
	AppendTo[freezingRenames,freezingRuleForObj];
	
	parentName=FMTSymbols`objectParentT@@lookupName;
	freezingRuleForParent=parentName->frozenT[parentName,id];
	AppendTo[freezingRenames,freezingRuleForParent];
	
	lhsName=FMTSymbols`objectLhsT@@lookupName;
	freezingRuleForLhs=lhsName->frozenT[lhsName,id];
	AppendTo[freezingRenames,freezingRuleForLhs];
	
	(* but return the freezing rule for the originally wrapped object: *)
	freezingRule=origName->frozenT[origName,id];
	
	freezingRule
]

SetAttributes[freezingReplacement,HoldAll]


applySingleFreezing[expression_,name_,freezingId_]:=
	If[freezingId>0,
		If[MemberQ[{FMTSymbols`objectT,FMTSymbols`objectParentT,FMTSymbols`objectLhsT},Head[name]],
			(* if the name to be replaced is on one of the above categories,
				perform replacement for all of them, as any of them can be present in expression: *)
			With[
				{
					objName=FMTSymbols`objectT@@name,
					parentName=FMTSymbols`objectParentT@@name,
					lhsName=FMTSymbols`objectLhsT@@name
				},
				
				expression/.{
					objName->frozenT[objName,freezingId],
					parentName->frozenT[parentName,freezingId],
					lhsName->frozenT[lhsName,freezingId]
				}
			],
			
			expression/.name->frozenT[name,freezingId]
		],
		expression
	]


(* Applies the freezing replacements within the given expression *)
applyCurrentFreezings[expression_,freezingRenames_Association]:=
	expression/.Normal[freezingRenames]


(* Creates new freeze and propagates domain for the given symbol *)
freezeSingleSymbol[name_,
	freezingRenames_,
	lastFreezingIds_,
	variableDomains_,
	newVariableDomains_,
	options_
]:=Module[
	{
		newNameRule,
		oldName,newName,
		newFullName, (* name with newNameRule replacement,
						this might be bigger than newName for aggregate accesses *)
		oldNameNoGrammar,newNameNoGrammar
	},
	
	newNameRule=freezingReplacement[name,lastFreezingIds,freezingRenames];
	oldName=First[newNameRule];
	newName=Last[newNameRule];
	
	(* propagate domains: *)
	
	newFullName=name/.newNameRule;
	
	(* as a matter of convenience for further processing, translate frozenT
		wrappers into subscripts (which is what unwrapFrozenAndGrammar would do as well): *)
	newName=unwrapFrozen[newName];
	newFullName=unwrapFrozen[newFullName];
	
	oldNameNoGrammar=FMTGrammar`unwrapGrammar[oldName,options];
	If[KeyExistsQ[variableDomains,oldNameNoGrammar],
		newNameNoGrammar=FMTGrammar`unwrapGrammar[newName,options];
		newVariableDomains[newNameNoGrammar]=variableDomains[oldNameNoGrammar]
	];
	
	(* in addition, check also full name (can contain aggregate accesses),
		if it happens to be different than newName *)
	If[newFullName=!=newName,
		oldNameNoGrammar=FMTGrammar`unwrapGrammar[name,options];
		If[KeyExistsQ[variableDomains,oldNameNoGrammar],
			newNameNoGrammar=FMTGrammar`unwrapGrammar[newFullName,options];
			newVariableDomains[newNameNoGrammar]=variableDomains[oldNameNoGrammar]
		]
	];
	
	newNameRule
]

SetAttributes[freezeSingleSymbol,HoldAll]


(* Freezes and replaces symbols in a single assignment operation *)
freezeAndReplaceSingleAssignment[a_,
	freezingRenames_,
	lastFreezingIds_,
	variableDomains_,
	newVariableDomains_,
	options_,
	OptionsPattern[{makeNewLhsFreeze->True,applyRhsFreezings->True}]
]:=Module[
	{
		(* a is assignmentT[objectLhsT[name,type],...] *)
		lhs=a[[1]],
		type=a[[1,2]],
		rhs=a[[2]],
		newNameRule,
		newLhsParts,newLhs,newRhs
	},
	
	(* Note: it is important here to replace lhs with newly created replacement name,
		while still using old replacements for rhs, to properly handle expressions like:
		x=x+1;
	*)
	
	If[OptionValue[applyRhsFreezings],
		newRhs=applyCurrentFreezings[rhs,freezingRenames],
		newRhs=rhs
	];
	
	If[OptionValue[makeNewLhsFreeze],
		(* apply freezings in all partT indices,
			they might refer to versioned variables,
			for example in a[[i+3]], where i is already versioned
		*)
		newLhsParts=ReplaceAll[lhs,
			FMTSymbols`partT[obj_,idx_]
				:>FMTSymbols`partT[obj,applyCurrentFreezings[idx,freezingRenames]]
		];
		
		(* create new freeze for lhs and propagate domain for the newly created freeze: *)
		newNameRule=freezeSingleSymbol[lhs,
			freezingRenames,lastFreezingIds,
			variableDomains,newVariableDomains,options
		];
		newLhs=newLhsParts/.newNameRule;
		
		(* additionally, add domain for the whole lhs, which might be different
			from the top-level object name, which is handled by freezeSingleSymbol *)
		newVariableDomains[newLhs]=type,
		
		(* do not create new freeze, just apply current freezings on lhs
			(this is used when the lhs freeze is known to already exist): *)
		newLhs=applyCurrentFreezings[lhs,freezingRenames]
	];
	
	{FMTSymbols`assignmentT[newLhs,newRhs]}
]

SetAttributes[freezeAndReplaceSingleAssignment,HoldAll]


End[];


EndPackage[];
