(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTSymbols`"];
Needs["FMTWorkarounds`"];


BeginPackage["FMTProver`"];


verifyProofs::usage="verifyProofs[sys,packageName,operationName,opProofStatements,variableDomains,enumMappings] Returns the result of verification of all proof conditions for the given operation.\n"<>
"The result has form {{index,description,wish,outcome},...}.";


Begin["`Private`"];


(* Helper, proves a single set of wishes against their knowns,
	the proof is performed by finding instances for specific wishes;
	this helper is used for proving wishes wrapped in
	neverT (in which case the given condition is considered to fail),
	and possibleT (in which case the given condition passes as reachable).
*)
findExamples[knowns_,wishes_,variableDomains_]:=Module[
	{
		examples
	},
	
	examples=Quiet[
		Check[
			FMTWorkarounds`findInstance[
				And@@Join[
					knowns,
					wishes,
					Element@@@Normal[variableDomains]
				]
			],
			
			(* If findInstance fails with error, it means that the problem
				is formulated in a way that is unsolvable using its methods.
				The most likely cause is reference to symbols that are
				not recognized in the model; such symbols are normally
				caught by the grammar, except for the case of assertions, assumptions
				and quantified statements,
				which do not restrict their expressions in any way.
			*)
			{{"Context"->"The problem is incomplete or too complex to solve."}}
		]
	];
	
	If[examples!={},
		First[examples],
		{}
	]
]


(* Helper, prooves a single set of wishes against their knowns,
	the proof is performed by finding instances for negated wishes;
	this helper is used for proving wishes wrapped in alwaysT
	(in which case the given condition is considered to fail).
*)
findContradictions[knowns_,wishes_,variableDomains_]:=
	findExamples[knowns,{Not[And@@wishes]},variableDomains]


(* Helper, checks whether expr1=>expr2,
	Example:
		expr1: a[[i]]\[Equal]5
		expr2: a[[i]]>0
		\[Rule] the match is True, because expr1 implies expr2
*)
expressionsMatchQ[vars1_,expr1_,vars2_,expr2_]:=Module[
	{
		(* expr1 rewritten with array accesses replaced with unique simple symbols: *)
		unifiedExpr1,
		
		(* expr2 rewritten with the use of variables from expr1,
			and then with array accesses replaced with unique simple symbols: *)
		unifiedExpr2,
		
		arrayAccesses,
		uniqueSymbols=<||>, (* <|arrayAccess\[Rule]uniqueSymbol,...|> *)
		uniqueSymbolsRules,
		result
	},
	
	arrayAccesses=Cases[{expr1,expr2},Inactive[Part][_,_],All];
	Scan[
		Function[aa, (* aa is array access expression *)
			uniqueSymbols[aa]=Unique[];
		],
		arrayAccesses
	];
	
	uniqueSymbolsRules=Normal[uniqueSymbols];
	unifiedExpr1=expr1/.uniqueSymbolsRules;
	unifiedExpr2=(expr2/.Thread[vars2->vars1])/.uniqueSymbolsRules;
	
	result=TrueQ[FullSimplify[unifiedExpr1\[Implies]unifiedExpr2]];
	
	Remove/@Values[uniqueSymbols];
	
	result
]

(* Helper, finds all ForAll or Exists statements within knowns that operate
	on (structurally) the same expression as the one given *)
(* TODO: Improvement idea:
	statement matching should not rely on structural equivalence alone,
	some tautologies can be reasonably detected as well, like
	a[[i]]\[Equal]0 => a[[i]]<10
*)
matchingQuantifiedStatements[knowns_,head_,vars_,expr_]:=Module[
	{
		result={}
	},
	
	Scan[
		Function[k, (* k is ForAll[vars,cond,expr] *)
			With[
				{
					kVars=k[[1]],
					kExpr=k[[3]]
				},
				
				If[expressionsMatchQ[kVars,kExpr,vars,expr],
					AppendTo[result,k]
				]
			]
		],
		Cases[knowns,_head]
	];
	
	result
]

(* helper, picks only the results for listed variables and negates them,
	for example x\[Rule]7 becomes x\[NotEqual]7 *)
pickNegatedResults[results_,vars_]:=Module[
	{
		varsList=If[Head[vars]===List,vars,{vars}],
		picked={}
	},
	
	Scan[
		Function[r, (* r is var\[Rule]value *)
			With[
				{
					var=r[[1]],
					value=r[[2]]
				},
				
				If[MemberQ[varsList,var],
					AppendTo[picked,Unequal[var,value]]
				]
			]
		],
		results
	];
	
	picked
]

(* Helper, removes those knowns from the set that involve quantified statements, at any level *)
deleteKnownsWithQuantStatements[knowns_]:=Map[
	Function[k, (* k is known *)
		If[Position[k,_ForAll|_Exists,All]=!={},
			Nothing, (* k involves some quantified statement,
				remove this known from the set *)
			k
		]
	],
	knowns
]

(* Helper, converts a single quantified statement into True/False,
	depending on whether it was possible to find a covering statement in the list of knowns *)
(* Note: there is a lot of intended duplication in algorithms between this function and
	validateDeduction/valudateInduction, but subtle differences between them
	mean that refactoring of this code could be too unreadable.
*)
resolveSingleQuantifiedStatement[knowns_,quantifiedStatement_,variableDomains_]:=Module[
	{
		(* quantifiedStatement is ForAll[vars,regions,expr] or Exists[...] *)
		quantVars=quantifiedStatement[[1]],
		quantRegion=quantifiedStatement[[2]],
		quantExpr=quantifiedStatement[[3]],
		
		knownsNoQuantStatements,
		matchingKnowns,
		matchingRegionCheck,
		atLeastOneRegionMatches,
		result={}
	},
	
	(* 1. find all matching quantified statements
		- that is, those that have the equivalent expression *)
	(* Improvement idea:
		Exists statements with singleton regions (that is, where
		there is only one solution) are equivalent to ForAll statements
		with everything else kept the same. This means that such Exists statements
		can be taken into account, too, when solving ForAlls.
		In general case, however, it can be necessary to involve a solver
		to decide whether the region is a singleton and it might be conditional
		on the value of other variables, if they are involved in region definition.
	*)
	knownsNoQuantStatements=deleteKnownsWithQuantStatements[knowns];
	matchingKnowns=matchingQuantifiedStatements[
		knowns,Head[quantifiedStatement],quantVars,quantExpr];
	
	(* 2. with ForAll statements, check
		assertT[Subset[deductionRegion,Union[matchingRegions]],
		which is done as neverT[deductionRegion&&Not[Union[matchingRegions]]],
		which is equivalent to neverT[deductionRegion&&Not[reg1]&&Not[reg2]&&...]
		
		with Exists statements, check if the deductionRegion covers
		one of the regions in matching knowns,
		which is done as assertT[Subset[matchingReg,deductionRegion]],
		which is done as neverT[matchingReg&&Not[deductionRegion]].
	*)
	If[Head[quantifiedStatement]===ForAll,
		matchingRegionCheck={quantRegion};
		Scan[
			Function[k, (* k is ForAll[vars,reg,expr] *)
				With[
					{
						vars=k[[1]],
						reg=k[[2]]
					},
					
					AppendTo[matchingRegionCheck,
						Not[reg/.Thread[vars->quantVars]]
					]
				]
			],
			matchingKnowns
		];
		
		result=findExamples[
			knownsNoQuantStatements,
			matchingRegionCheck,
			Append[
				variableDomains,
				Association[Thread[quantVars->Integers]]
			]
		];
		
		(* Empty result here means that there were no contradictions
			and the quantified statement can be considered True *)
		result=={},
		
		(* Exists \[Rule] check exact matching for each found known separately *)
		atLeastOneRegionMatches=False;
		Scan[
			Function[k, (* k is Exists[vars,reg,expr] *)
				If[atLeastOneRegionMatches==False, (* skip the test if match already found *)
					With[
						{
							vars=k[[1]],
							reg=k[[2]]
						},
						
						(* finding matching between regions can be done
							by checking if Subset[reg,inductionRegion] *)
						matchingRegionCheck={
							(reg/.Thread[vars->quantVars])&&Not[quantRegion]
						};
						
						result=findExamples[
							knownsNoQuantStatements,
							matchingRegionCheck,
							Append[
								variableDomains,
								Association[Thread[quantVars->Integers]]
							]
						];
						
						If[result=={},
							atLeastOneRegionMatches=True,
							
							(* the quantified region does not fully cover
								the candidate known statement;
								that is, existing quantified statements of type Exists
								operate on wider regions than the one used in this statement,
								so it is not possible to resolve it
							*)
							Null
						]
					]
				]
			],
			matchingKnowns
		];
		
		atLeastOneRegionMatches
	]
]

resolveQuantifiedStatements[knowns_,expression_,variableDomains_]:=With[
	{
		quantPositions=Position[expression,_ForAll|_Exists,All]
	},
	
	MapAt[
		Function[expr,
			resolveSingleQuantifiedStatement[knowns,expr,variableDomains]
		],
		expression,
		quantPositions
	]
]

validateDeduction[knowns_,quantifiedStatement_,instance_,variableDomains_]:=Module[
	{
		(* quantifiedStatement is ForAll[vars,regions,expr] or Exists[...] *)
		deductionVars=quantifiedStatement[[1]],
		deductionRegion=quantifiedStatement[[2]],
		deductionExpr=quantifiedStatement[[3]],
		instanceList,
		instanceInRegion,
		knownsNoQuantStatements,
		matchingKnowns,
		matchingRegionCheck,
		atLeastOneRegionMatches,
		result
	},
	
	instanceList=If[Head[instance]===List,instance,{instance}];
	
	(* applying instance replacement (var\[Rule]value) on the region allows to
		perform the most basic check of the deduction validity: *)
	instanceInRegion={deductionRegion/.instanceList};
	
	(* it is possible for the instanceInRegion to be immediately resolved to False,
		for example when the instance is outside of the region and this does not
		depend on any other variables; in this case no further checks are necessary
	*)
	If[instanceInRegion=!={False},
	
		knownsNoQuantStatements=deleteKnownsWithQuantStatements[knowns];
		
		(* 1. check the equivalent of assertT[deductionRegion/.instance]
			taking into account all other knowns
			- reuse findExamples/findContradictions for this, just as for regular assertT *)
		
		If[Head[quantifiedStatement]===ForAll,
			(* a simple call to findContradictions works with ForAll,
				this is the test for whether the instance is covered by the ForAll region,
				and the replacement in instanceInRegion above did not immediately resolve to False: *)
			result=findContradictions[knowns,instanceInRegion,variableDomains],
			
			(* otherwise (for Exists), the check is compound:
				- the instance needs to be covered by the Exists region, and
				- there is nothing left in the region
			*)
			result=Join[
				findContradictions[knowns,instanceInRegion,variableDomains],
				findExamples[
					knowns,
					deductionRegion&&Thread[deductionVars!=deductionVars/.instanceList],
					Append[
						variableDomains,
						Association[Thread[deductionVars->Integers]]
					]
				]
			]
		];
		
		If[result=={},
			(* no contradictions were found for the instance *)
			
			(* 2. find all matching quantified statements
				- that is, those that have the equivalent expression *)
			(* Improvement idea:
				Exists statements with singleton regions (that is, where
				there is only one solution) are equivalent to ForAll statements
				with everything else kept the same. This means that such Exists statements
				can be taken into account, too, when solving ForAlls.
				In general case, however, it can be necessary to involve a solver
				to decide whether the region is a singleton and it might be conditional
				on the value of other variables, if they are involved in region definition.
			*)
			matchingKnowns=matchingQuantifiedStatements[
				knowns,Head[quantifiedStatement],deductionVars,deductionExpr];
			
			(* 3. with ForAll statements, check
				assertT[Subset[deductionRegion,Union[matchingRegions]],
				which is done as neverT[deductionRegion&&Not[Union[matchingRegions]]],
				which is equivalent to neverT[deductionRegion&&Not[reg1]&&Not[reg2]&&...]
				
				with Exists statements, check if the deductionRegion covers
				one of the regions in matching knowns,
				which is done as assertT[Subset[matchingReg,deductionRegion]],
				which is done as neverT[matchingReg&&Not[deductionRegion]].
			*)
			If[Head[quantifiedStatement]===ForAll,
				matchingRegionCheck={deductionRegion};
				Scan[
					Function[k, (* k is ForAll[vars,reg,expr] *)
						With[
							{
								vars=k[[1]],
								reg=k[[2]]
							},
							
							AppendTo[matchingRegionCheck,
								Not[reg/.Thread[vars->deductionVars]]
							]
						]
					],
					matchingKnowns
				];
				
				result=findExamples[
					knownsNoQuantStatements,
					matchingRegionCheck,
					Append[
						variableDomains,
						Association[Thread[deductionVars->Integers]]
					]
				];
				
				If[result!={},
					result={"Coverage"->"Deduction region is not covered by existing knowns."}
				],
				
				(* Exists \[Rule] check exact matching for each found known separately *)
				atLeastOneRegionMatches=False;
				Scan[
					Function[k, (* k is Exists[vars,reg,expr] *)
						If[atLeastOneRegionMatches==False, (* skip the test if match already found *)
							With[
								{
									vars=k[[1]],
									reg=k[[2]]
								},
								
								(* finding exact matching between regions can be done
									by checking if Subset[reg,deductionRegion] *)
								matchingRegionCheck={
									(reg/.Thread[vars->deductionVars])&&Not[deductionRegion]
								};
								
								result=findExamples[
									knownsNoQuantStatements,
									matchingRegionCheck,
									Append[
										variableDomains,
										Association[Thread[deductionVars->Integers]]
									]
								];
								
								If[result=={},
									(* check if the Exists region is a singleton.
										This is done by solving the region once and checking
										whether there is any other solution after
										excluding the first solution from the region.
									*)
									With[
										{
											domainsWithRegionVars=Append[
												variableDomains,
												Association[Thread[vars->Integers]]
											]
										},
										
										(* try to find first set of solutions: *)
										result=findExamples[
											knownsNoQuantStatements,
											{reg},
											domainsWithRegionVars
										];
										
										(* try to find another example,
											excluding the results obtained in the first try above: *)
										result=findExamples[
											knownsNoQuantStatements,
											{reg&&And@@pickNegatedResults[result,vars]},
											domainsWithRegionVars
										];
										
										If[result!={},
											result={"Coverage"->"Deduction region is not a singleton."}
										]
									];
									
									If[result=={},
										(* if there are no more results,
											then the region is a singleton
											- that is, it has only one possible value,
											which allows to deduce from Exists
										*)
										atLeastOneRegionMatches=True
									],
									
									(* the deduction region does not fully cover
										the candidate known statement;
										that is, existing quantified statements of type Exists
										operate on wider regions than the one used for deduction,
										so it is not possible to deduce anything from them
										in the narrower scope of the deduction region
									*)
									result={"Coverage"->"Deduction region does not cover the existing knowns."}
								]
							]
						]
					],
					matchingKnowns
				];
				
				If[!atLeastOneRegionMatches,
					If[result=={},
						result={"Coverage"->"No matching quantified statement was found."}
					]
				]
			],
			
			(* contradiction in the instance is best presented as the instance itself *)
			result=instanceList
		],
		
		(* contradiction in the instance was detected immediately,
			and here, too, is best presented as the instance itself *)
		result=instanceList
	];
	
	result
]

validateInduction[knowns_,quantifiedStatement_,extension_,variableDomains_]:=Module[
	{
		(* quantifiedStatement is ForAll[vars,region,expr] or Exists[...] *)
		inductionVars=quantifiedStatement[[1]],
		inductionRegion=quantifiedStatement[[2]],
		inductionExpr=quantifiedStatement[[3]],
		extensionList,
		(*instanceInRegion,*)
		knownsNoQuantStatements,
		matchingKnowns,
		matchingRegionCheck,
		atLeastOneRegionMatches,
		result
	},
	
	(* The validation of inductions is performed in two phases:
		1. check that the quantified statement is indeed covered by the existing knowns,
			in the case of ForAll, this check is the same as with validateDeduction,
			except that it is possible to induce from empty regions (these are always valid);
			in the case of Exists it is also different as it checks whether the given
			quantified statement encompasses at least one of the knowns,
		2. check the statement at the extension point.
	*)
	
	extensionList=If[Head[quantifiedStatement]===ForAll,
		extensionList=extension,
		
		(* in the case of Exists, extension has the form Not[...],
			or, when the region is \[EmptySet], the extension does not have any wrapper *)
		If[inductionRegion=!=FMTSymbols`\[EmptySet],
			extensionList=extension[[1]],
			extensionList=extension
		]
	];
	extensionList=If[Head[extensionList]===List,extensionList,{extensionList}];
	knownsNoQuantStatements=deleteKnownsWithQuantStatements[knowns];
	
	(* Phase 1. - check the coverage with respect to the existing statements. *)
	
	(* find all matching quantified statements
		- that is, those that have the equivalent expression *)
	
	matchingKnowns=matchingQuantifiedStatements[
		knowns,Head[quantifiedStatement],inductionVars,inductionExpr];
	
	result={}; (* in validateDeduction above, at this point result has some value
					from the instance check that was computed just before;
					here, for consistency, an empty list is set
					to emulate success from previous stages *)
	(* if induction region is an empty set, consider the region as valid,
		this is allowed for both ForAll and Exists *)
	If[inductionRegion=!=FMTSymbols`\[EmptySet],
		If[Head[quantifiedStatement]===ForAll,
			matchingRegionCheck={inductionRegion};
			Scan[
				Function[k, (* k is ForAll[vars,reg,expr] *)
					With[
						{
							vars=k[[1]],
							reg=k[[2]]
						},
						
						AppendTo[matchingRegionCheck,
							Not[reg/.Thread[vars->inductionVars]]
						]
					]
				],
				matchingKnowns
			];
			
			result=findExamples[
				knownsNoQuantStatements,
				matchingRegionCheck,
				Append[
					variableDomains,
					Association[Thread[inductionVars->Integers]]
				]
			];
			
			If[result!={},
				(* if result is already {"..."\[Rule]"..."}, then leave it as is,
					these strings contain human-readable description of the problem;
					otherwise blame the quantified statement for the failed check *)
				If[!MatchQ[result,List[_String->_String]],
					result={"Coverage"->"Induction region is not covered by existing knowns."}
				]
			],
		
			(* Exists \[Rule] check matching for each found known separately *)
			atLeastOneRegionMatches=False;
			Scan[
				Function[k, (* k is Exists[vars,reg,expr] *)
					If[atLeastOneRegionMatches==False, (* skip the test if match already found *)
						With[
							{
								vars=k[[1]],
								reg=k[[2]]
							},
							
							(* finding matching between regions can be done
								by checking if Subset[reg,inductionRegion] *)
							matchingRegionCheck={
								(reg/.Thread[vars->inductionVars])&&Not[inductionRegion]
							};
							
							result=findExamples[
								knownsNoQuantStatements,
								matchingRegionCheck,
								Append[
									variableDomains,
									Association[Thread[inductionVars->Integers]]
								]
							];
							
							If[result=={},
								atLeastOneRegionMatches=True,
								
								(* the induction region does not fully cover
									the candidate known statement;
									that is, existing quantified statements of type Exists
									operate on wider regions than the one used for induction,
									so it is not possible to induce anything from them
								*)
								result={"Coverage"->"Induction region does not cover the existing knowns."}
							]
						]
					]
				],
				matchingKnowns
			];
			
			If[!atLeastOneRegionMatches,
				If[result=={},
					result={"Coverage"->"No matching quantified statement was found."}
				]
			]
		]
	];
	
	If[result=={},
		(* Phase 2. - check the extension point. *)
		
		If[Head[quantifiedStatement]===ForAll,
			(* In the case of ForAll, check that the expression is True
				in the point of extension, to justify larger region in the newly created statement;
				this check is being done as assertT[expression] at the extension point.
			*)
			result=findContradictions[
				knownsNoQuantStatements,
				{inductionExpr/.extensionList},
				variableDomains
			],
			
			(* In the case of Exists, check that the expression is False
				in the point of extension, to justify smaller region in the newly created statement;
				the exception is Exists with inductionRegion of \[EmptySet], which
				should check whether the expression is True, just like with ForAll
			*)
			If[inductionRegion=!=FMTSymbols`\[EmptySet],
				(* Exists[vars,region,Not[extension]] *)
				result=findExamples[
					knownsNoQuantStatements,
					{inductionExpr/.extensionList},
					variableDomains
				],
				
				(* Exists[vars,\[EmptySet],extension] *)
				result=findContradictions[
					knownsNoQuantStatements,
					{inductionExpr/.extensionList},
					variableDomains
				]
			]
		]
	];
	
	result
]


(* Helper, reverts numeric mappings for enum types
	(that is, changes key\[Rule]value into value\[Rule]key mappings *)
revertMappings[mappings_]:=Map[
	Function[m, (* m is {k\[Rule]v,...} *)
		Association[Map[Reverse,m]]
	],
	mappings
]

(* Helper, applies reverse enum mappings to counterexamples, to recover original enumeration names
	on values found as solution instances in the Integers domain *)
applyReverseEnumMappings[singleResult_,originalDomains_,enumMappings_]:=With[
	{
		reverseEnumMappings=revertMappings[enumMappings]
	},
	
	Map[
		Function[ce, (* ce is name\[Rule]value *)
			With[
				{
					name=ce[[1]],
					value=ce[[2]]
				},
				
				If[KeyExistsQ[originalDomains,name],
					With[
						{
							domainOrType=originalDomains[name]
						},
						
						If[domainOrType=!=Integers&&domainOrType=!=Booleans,
							If[KeyExistsQ[reverseEnumMappings[domainOrType],value],
								name->reverseEnumMappings[domainOrType][value],
								name->Indeterminate
							],
							
							(* the type is Integers or Booleans, no restoration *)
							name->value
						]
					],
					
					(* value has no explicit domain assigned,
						(it is initializedT[var] tag, which is implicitly Boolean),
						\[Rule] no restoration *)
					name->value
				]
			]
		],
		singleResult
	]
]


(* Helpers, check whether the list of positions {{a,b,c,...},...}
	contains any entry that has no prefix in the list of candidate prefix positions
*)
isPrefixQ[pos1_,pos2_]:=With[
	{
		len1=Length[pos1],
		len2=Length[pos2]
	},
	
	len1<len2&&pos1==Take[pos2,len1]
]

hasPrefixQ[pos_,prefixCandidates_]:=Or@@Map[
	Function[pc,
		isPrefixQ[pc,pos]
	],
	prefixCandidates
]

hasPositionWithoutPrefixQ[positions_,prefixCandidates_]:=Or@@Map[
	Function[pos,
		Not[hasPrefixQ[pos,prefixCandidates]]
	],
	positions
]


(* Helper, performs the following:
	- drops all statements that involve ForAll, Exists, whole arrays or whole structures
	- applies enum mappings
	The result is a list of statements on Integers and Booleans only.
*)
pruneKnowns[knowns_,
	retainQuantifiedStatements_,
	variableDomains_Association,
	enumReplacements_
]:=Module[
	{
		simpleVars={}, (* Integers, Booleans and enum variable names only *)
		otherVars={},
		result={}
	},
	
	(* identify simple variables (Integers, Booleans and enums) and others *)
	Scan[
		Function[v, (* v is name\[Rule]domain *)
			With[
				{
					name=v[[1]],
					domain=v[[2]]
				},
				
				If[MatchQ[domain,Integers|Booleans|_FMTSymbols`enum],
					AppendTo[simpleVars,name],
					AppendTo[otherVars,name]
				]
			]
		],
		Normal[variableDomains]
	];
	
	(* Variables can include names that are partially overlapping, for example
		x and x[[7]] are two names that can appear in knowns,
		and only x[[7]] might be simple (but does not have to).
		For the purpose of built-in proofs, only simple variables are accepted,
		but since variable names can overlap, they have to be detected in a smart way;
		below both simple and other variables are identified by position
		and positions are analyzed for overlaps;
		variables from the "other" set that do not overlap with simple variables
		indicate that the whole known should be pruned.
		
		For example, in the case of x and x[[7]], x[[7]] has a shorter position,
		because it is matched as a more complete construct that exists higher
		in the expression tree, so the position of x[[7]] is a prefix of x's;
		the presence of x without respective prefix that would be matched by x[[7]]
		means that x was used as a whole object and that excludes the candidate known
		from the list of knowns that are taken into account by the solver.
	*)
	Scan[
		Function[k, (* k is a single known (but possibly a complex expression) *)
			If[MemberQ[k,_ForAll|_Exists,All],
				If[retainQuantifiedStatements,
					(* quantified statements are retained only if explicitly requested *)
					AppendTo[result,k/.enumReplacements]
				],
				
				(* non-quantified expressions: *)
				With[
					{
						simplePositions=Flatten[
							Map[
								Function[v, (* v is name of simple variable *)
									Position[k,v,All]
								],
								simpleVars
							],
							1
						],
						otherPositions=Flatten[
							Map[
								Function[v, (* v is name of other variable *)
									Position[k,v,All]
								],
								otherVars
							],
							1
						]
					},
					
					(* known is accepted if it does not contain other vars that
						do not have overlapping (as prefixes) simple vars: *)
					If[Not[hasPositionWithoutPrefixQ[otherPositions,simplePositions]],
						AppendTo[result,k/.enumReplacements]
					]
				]
			]
		],
		knowns
	];
	
	
	result
]

SetAttributes[pruneKnowns,HoldFirst]


(* Accepts the list of proof statements from FMTProofs`proofStatements
	(a single invocation is for a single operation)
	and returns the list {{index,{description,wish},outcome},...}, where
	index is a point number in the given operations body,
	description is a human-readable "correctness headline",
	wish is a statement that needs to be true in the given part of operation,
	outcome is one of:
		provenT[] (if the given wish is satisfied),
		impossibleT[] (if the possibleT wish is unreachable),
		failedT[contradictions] (if the given wish is not satisfied,
								contradictions is an example list of variable assignments (rules),
								for which the given wish fails)
*)
verifyProofs[sys_,
	packageName_String,
	qOperationName_Symbol,
	opProofStatements_,
	variableDomains_Association,
	enumMappings_
]:=Module[
	{
		mappedDomains,
		enumTypesUsed,
		enumReplacements,
		domains, (* domains for variables and their initialization flags *)
		singleProofResult,
		quantResolvedWish, (* temporary, for storing wishes with resolved quantified statements *)
		temp,
		results={} (* {{index,wish,outcome},...} *)
	},
	
	(* variableDomains contains domains for all symbols and constructs in opProofStatements,
		including array accesses and structure fields *)
	
	(* prepare enumeration mappings to both proof statements and variable domains,
		but keep the original domain information to enable reverse mapping on found counterexamples *)
	
	enumTypesUsed=Keys[enumMappings];
	enumReplacements=Join@@Values[enumMappings];
	mappedDomains=variableDomains/.Thread[enumTypesUsed->Integers];
	
	(* in addition to all domains that were assigned for variables as used in operation body,
		assign also domains for tagged values like initializedT[var] *)
	
	temp=Cases[opProofStatements,
		in_FMTSymbols`initializedT->Rule[in,Booleans],
		All
	];
	temp=Union[temp];
	domains=Append[mappedDomains,temp];
	
	(* add also domains of ghost variables, if there are any *)
	
	AppendTo[domains,FMTSystem`fmtGetGhostDomains[sys]];
	
	Scan[
		Function[proof, (* proof is {index,knowns,{{description,wish},...}} *)
			With[
				{
					index=proof[[1]],
					
					(* apply enum mappings and retain quantified statements,
						but prune other knowns *)
					quantifiedKnowns=pruneKnowns[proof[[2]],True,domains,enumReplacements],
					
					(* prune knowns and retain only those that can be solved by findExample
						(that is, Integers/Booleans and already mapped enums,
						without quantified statements): *)
					prunedKnowns=pruneKnowns[proof[[2]],False,domains,enumReplacements],
					
					wishes=proof[[3]]
				},
				
				(* Note:
					for the given index, all wishes are processes separately;
					as an optimization strategy (to be checked!)
					consider grouping wishes of the same type
					(that is: group by neverT, alwaysT and possibleT)
				*)
				
				Scan[
					Function[describedWish, (* {description,wish} *)
						With[
							{
								description=describedWish[[1]],
								originalWish=describedWish[[2]],
								mappedWish=describedWish[[2]]/.enumReplacements
							},
							
							Switch[Head[originalWish],
								FMTSymbols`alwaysT,
									If[originalWish===FMTSymbols`alwaysT[False],
										(* special case of wish that always fails *)
										singleProofResult=FMTSymbols`failedT[],
										
										(* regular alwaysT wish *)
										quantResolvedWish=resolveQuantifiedStatements[
											quantifiedKnowns,
											List@@mappedWish,
											domains
										];
										temp=findContradictions[
											prunedKnowns,
											quantResolvedWish,
											domains
										];
										If[temp!={},
											(* an example contradiction was found \[Rule]
												condition failed *)
											singleProofResult=applyReverseEnumMappings[
												FMTSymbols`failedT@@temp,variableDomains,enumMappings],
											
											(* no contradictions \[Rule] condition proved correct *)
											singleProofResult=FMTSymbols`provenT[]
										]
									],
								
								FMTSymbols`neverT,
									quantResolvedWish=resolveQuantifiedStatements[
										quantifiedKnowns,
										List@@mappedWish,
										domains
									];
									temp=findExamples[
										prunedKnowns,
										quantResolvedWish,
										domains
									];
									If[temp!={},
										(* an example was found for what should never happen \[Rule]
											condition failed *)
										singleProofResult=applyReverseEnumMappings[
											FMTSymbols`failedT@@temp,variableDomains,enumMappings],
										
										(* no contradictions \[Rule] condition proved correct *)
										singleProofResult=FMTSymbols`provenT[]
									],
								
								FMTSymbols`possibleT,
									quantResolvedWish=resolveQuantifiedStatements[
										quantifiedKnowns,
										List@@mappedWish,
										domains
									];
									temp=findExamples[
										prunedKnowns,
										quantResolvedWish,
										domains
									];
									If[temp!={},
										(* an example was found \[Rule] condition proved correct *)
										singleProofResult=FMTSymbols`provenT[],
										
										(* no example found for what should be possible \[Rule]
											condition failed *)
										singleProofResult=FMTSymbols`impossibleT[]
									],
								
								FMTSymbols`deductionT,
									temp=validateDeduction[
										quantifiedKnowns,
										mappedWish[[1]], (* ForAll[...] or Exists[...] *)
										mappedWish[[2]], (* instance *)
										domains
									];
									If[temp===Null,
										(* deduction failed, there was no appropriate known to rely on *)
										singleProofResult=FMTSymbols`failedT[],
										
										If[temp!={},
											(* deduction failed, because instance
												does not fit the condition *)
											singleProofResult=applyReverseEnumMappings[
												FMTSymbols`failedT@@temp,variableDomains,enumMappings],
											
											(* no contradictions \[Rule] deduction was validated *)
											singleProofResult=FMTSymbols`provenT[]
										]
									],
								
								FMTSymbols`inductionT,
									temp=validateInduction[
										quantifiedKnowns,
										mappedWish[[1]], (* ForAll[...] or Exists[...] *)
										mappedWish[[2]], (* extension *)
										domains
									];
									If[temp!={},
										(* an example contradiction was found \[Rule]
											extension failed *)
										singleProofResult=applyReverseEnumMappings[
											FMTSymbols`failedT@@temp,variableDomains,enumMappings],
										
										(* no contradictions \[Rule] extension proved correct *)
										singleProofResult=FMTSymbols`provenT[]
									]
							];
							
							AppendTo[results,
								{
									index,
									{
										description,
										originalWish
									},
									singleProofResult
									
									(* Note: it might be possible to append this instead, as the last element of this list:
										
										singleProofResult/.reverseFreezings
										
										but then the same object can appear more than once in the list of solutions
										for a failing counterexample, which is misleading.
										By leaving the variable names in their frozen forms, it is also possible
										to track the evolution of the value from the beginning of the operation
										(where all object have their original symbol names) up to the point
										of proof (where objects have their frozen replacements).
										The price for this is a bit more cryptic report (with frozen names),
										but it seem to be worth the effort.
										
										The reverseFreezings values are returned as the last component of each element
										in the list from operationProofStatements, in case they are useful.
									*)
								}
							]
						]
					],
					wishes
				]
			]
		],
		opProofStatements
	];
	
	results
]


End[];


EndPackage[];
