(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


BeginPackage["FMTSymbols`"];


systemDefinitionTags={
	array, (* for defining array types *)
	base, (* to name the base type in type definitions (now only in number),
		and (internally) to denote a simple type category (boolean, integer or enum) *)
	enum, (* for defining enumeration types *)
	in, (* input parameter *)
	inOut, (* input/output parameter *)
	invariant, (* type invariant in type definition (now only in number) *)
	number, (* for defining numeric types *)
	out, (* output parameter *)
	post, (* operation's post-condition *)
	pre, (* operation's pre-condition *)
	structure, (* for defining structural types *)
	
	\[EmptySet] (* for denoting empty sets *)
};


grammarTags={
	actualParamsT, (* helper tag to gather together the list of call's actual parameters *)
	alternativesT, (* Alternatives *)
	andT, (* And *)
	assertT, (* Assert *)
	assignmentT, (* Set *)
	Assume, (* Note: Assume is not a standard Wolfram symbol,
				but is needed for consistency with Assert *)
	assumeT, (* Assume *)
	booleanT, (* built-in boolean type, also as alias boolean *)
	callT, (* operation call *)
	Comment, (* Note: Comment, for user-injected code comments,
				is not a standard Wolfram symbol,
				but is spelled this way for consistency with other constructs *)
	commentT, (* Comment *)
	compoundT, (* CompoundExpression *)
	Deduce, (* Note: Deduce is not a standard Wolfram symbol,
				but is needed for consistency with Assert and Assume *)
	deduceT, (* Deduce *)
	defaultT, (* default Switch branch *)
	divT, (* division, sane replacement for Times[p,Power[q,-1]] *)
	divByZeroT, (* division by literal zero detected, this is intended to be a grammar error *)
	dotT, (* Dot *)
	equalT, (* Equal, \[Equal] *)
	exprT, (* wrapper for expression results (including reductions from single objects and literals *)
	forT, (* intermediary symbol, used in the middle of For\[Rule]whileT preprocessing *)
	formalParamsT, (* helper tag to gather together the list of operation's formal parameters *)
	GhostCapture, (* Note: GhostCapture is not a standard Wolfram symbol,
					but fits in the set of other verification-related symbols *)
	ghostCaptureT, (* GhostCapture *)
	GhostRecall,
	ghostRecallT, (* GhostRecall *)
	greaterT, (* Greater, > *)
	greaterEqualT, (* GreaterEqual, \[GreaterEqual] *)
	highlightedT, (* wrapper for highlighting numbered expression parts *)
	Induce, (* Note: Deduce is not a standard Wolfram symbol,
				but is needed for consistency with Assert, Assume and Deduce *)
	induceT, (* Induce *)
	ifT, (* If *)
	ifUnrolledT, (* If, which was injected from loop unrolling, no alternative branch expected *)
	integerT, (* built-in base integer type, also as alias integer *)
	lessT, (* Less, < *)
	lessEqualT, (* LessEqual, \[LessEqual] *)
	literalT, (* wrapper for integer and enum literals *)
	lockedLiteralT, (* wrapper for literals that are not further transformed into exprT (they stay as literals) *)
	LoopInvariant, (* Note: LoopInvariant and LoopVariant are not standard Wolfram symbols,
						but are uppercase for consistency with Assert, Assume, Deduce and Induce *)
	loopInvariantT, (* LoopInvariant *)
	LoopVariant,
	loopVariantT,
	matchingParamT, (* marker tag for matching call parameters *)
	negT, (* negation, to capture Times[x,-1] as -x *)
	notT, (* Not *)
	numberedT, (* optional wrapper for numbering statements *)
	objectT, (* wrapper for object references (rhs) *)
	objectLhsT, (* wrapper for object references (lhs in assignments) *)
	objectParentT, (* wrapper for objects that are containers for others (arrays or structures) *)
	orT, (* Or *)
	partT, (* Part *)
	plusT, (* Plus, + *)
	statementT, (* any statement, used to represent either an empty statement or a result of reduction *)
	switchT, (* Switch *)
	Synthesize, (* code-synthesis marker, not translated into later grammar tags *)
	partToChangeT, (* temporary helper for selective replacements *)
	timesT, (* Times, * *)
	unequalT, (* Unequal, \[NotEqual] *)
	whileT (* While *)
};

verificationTags={
	initializedT, (* tag stating whether the given local variable was initialized or not *)
	alwaysT, (* wraps conditions that should be true for all possible executions *)
	neverT, (* wraps conditions that should not be true for any possible execution *)
	possibleT, (* wraps conditions that should be possible to achieve (reachable) *)
	deductionT, (* wraps deduction from ForAll|Exists quantified statements *)
	inductionT, (* wraps induction of new quantified statements *)
	oldT (* wraps symbols (or whole expressions) to indicate that
			the intended value of the symbol (or expression)
			is that from the beginning of the given operation *)
};

proverTags={
	provenT, (* represents successful proof of some condition *)
	impossibleT, (* represents unreachability of some condition *)
	failedT (* represents contradiction of some condition,
				wraps the list of contradicting variable assignments *)
};

(* List of grammar tags that indicate the presence of simple-typed, non-object expression *)
simpleExpressionTags={
	andT,
	divT,
	equalT,
	greaterT,
	greaterEqualT,
	lessT,
	lessEqualT,
	literalT,
	negT,
	notT,
	orT,
	plusT,
	timesT,
	unequalT
};

standardTypes={
	booleanT,
	integerT
};


(* User-friendly aliases for standard data types. *)
boolean=booleanT;
integer=integerT;

(* User-friendly aliases for verification tags (only those used by the user). *)
old=oldT;


codeGeneratorTags={
	fileNamesFromPkgName,
	nameFromSymbol,
	fileHeaderComment,
	integerTargetType,
	arrayCopyOperation
};


EndPackage[];
