(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


BeginPackage["FMTWorkarounds`"];


findInstance::usage="findInstance[statements] finds instances for the given problem, which should consist entries of the form Element[variable,domain].\n"<>
"This operation accepts only Booleans and Integers as domains.";


Begin["`Private`"];


(* findInstance is a replacement for Wolfram's FindInstance
	with work-arounds for the following issues:
	1. FindInstance cannot cope with mixed-domain (Integers/Booleans) problems.
		This is solved by identifying all Boolean variables and replacing them
		by all True/False combinations. This is a brute-force approach, but since
		findInstance is used only in contexts where at least one solution is sufficient
		to prove the goal, the loop is broken after first variable replacement combination
		that leads to non-empty result. This usually happens very quickly.
	2. FindInstance occasionally hangs (or rather livelocks with insane amount of consumed memory)
		when solving expressions containing IntegerPart[x] symbols (IntegerPart is used
		by default for intermediate results of integer divide operations).
		The solution is to identify all IntegerPart[x] subexpressions and replace them
		with additional integer variables that are separate injected (And'ed) to the complete
		expression as temp=x terms.
		These additional variables are later removed from the result set.
	3. FindInstance cannot work with variables that are parts of each other,
		for example x and Subscript[x,1] - these are disctinct variables in proof statements,
		but FindInstance breaks down with them.
		The workaround (in helper findInstanceUniqueVars) is to replace all variables
		from the list with unique symbols, and then do the reverse replacement in solutions.
	4. FindInstance cannot work with empty problem: FindInstance[False,{},Integers],
		but it is convenient to support it.
*)

(* Helper, solve issue 3. above *)
findInstanceUniqueVars[expr_,vars_,domain_]:=Module[
	{
		uniqueVars={}, (* {unique1\[Rule]var1,...} *)
		results={}
	},
	
	If[expr=!=False,
		
		(* 1. prepare replacements with unique variables for all variables in the list *)
		Scan[
			Function[v,
				With[
					{
						uniqueVar=Unique["u"]
					},
					
					AppendTo[uniqueVars,v->uniqueVar]
				]
			],
			vars
		];
		
		(* 2. replace all variables with unique names and solve replaced expression *)
		If[uniqueVars!={},
			results=FindInstance[expr/.uniqueVars,uniqueVars[[All,2]],domain],
				
			(* conventionally, empty problem has no solutions;
				at this level, expr has the form And[...] *)
			If[Length[expr]!=0,
				(* this message simulates the error within FindInstance;
					it is detected and translated into a failure report at higher level *)
				Message[FindInstance::exvar]
			]
		];
		
		(* 3. reverse-replace in solutions to revert the original variable names,
			and get rid of the temporary unique symbols *)
		results=results/.(Reverse/@uniqueVars);
		Scan[Remove,uniqueVars[[All,2]]]
	];
	
	results
]

(* Helper, returns an equivalent implementation for resultVar==IntegerPart[p/q] *)
makeIntegerPartFormula[p_,q_,resultVar_]:=
	If[(p>=0&&q>0)||(p<0&&q<0),
		resultVar>(p-q)/q&&resultVar<=p/q,
		resultVar>=p/q&&resultVar<(p+q)/q
	]

(* Main workaround function *)
findInstance[expr_]:=Module[
	{
		varBools, (* all Boolean variables *)
		varInts, (* all Integer variables *)
		noDomainsExpr, (* expr without domain information var\[Element]Domain *)
		boolTuples, (* brute-force Boolean replacement tuples *)
		ipCases, (* all IntegerPart[x] subexpressions *)
		ipReplacements={}, (* {IntegerPart[x]\[Rule]tempVar,...} *)
		ipEquivalents={}, (* statements that compute the equivalent of temp==IntegerPart[x] *)
		ipVars={}, (* names of variables for IntegerPart replacements: {tempVar,...} *)
		tempExpr, (* single candidate expr for Boolean replacement *)
		tempResults, (* single result for Boolean replacement *)
		results
	},
	
	varBools=Cases[expr,Element[v_,Booleans]->v];
	varInts=Cases[expr,Element[v_,Integers]->v];
	
	noDomainsExpr=DeleteCases[expr,_Element,{1}];
	
	(* replace IntegerPart[x] with additional variables temp\[Equal]equivalentFormula[x]
		and add new temp vars to the overall list of integer variables *)
	ipCases=Cases[expr,_IntegerPart,All];
	Scan[
		Function[ip, (* ip is IntegerPart[ex] *)
			With[
				{
					tempVar=Unique["temp"],
					
					(* pq is {p,q} from IntegerPart[p/q],
						there are two possible representations for p/q: *)
					pq=ip[[1]]/.
						{
							(* q was an integer literal and 1/q is a Rational: *)
							Times[Rational[1,q_],p_]->{p,q},
							
							(* q was an expression and 1/q is represented with Power *)
							Times[p_,Power[q_,-1]]->{p,q}
						}
				},
				
				AppendTo[ipReplacements,ip->tempVar];
				AppendTo[ipEquivalents,makeIntegerPartFormula[pq[[1]],pq[[2]],tempVar]];
				AppendTo[ipVars,tempVar]
			]
		],
		ipCases
	];
	noDomainsExpr=And[noDomainsExpr/.ipReplacements,And@@ipEquivalents];
	varInts=Union[varInts,ipVars];
	
	If[varBools=={},
		(* this is Integers-only problem *)
		results=findInstanceUniqueVars[noDomainsExpr,varInts,Integers],
		
		If[varInts=={},
			(* this is Booleans-only problem *)
			results=findInstanceUniqueVars[noDomainsExpr,varBools,Booleans],
			
			(* mixed domains *)
			
			results={};
			
			boolTuples=Map[
				Function[tuple,
					Thread[varBools->tuple]
				],
				Tuples[{False,True},Length[varBools]]
			];
			
			(* as a matter of optimization,
				allow to break in the middle of iteration if any solution is found: *)
			Catch[
				Scan[
					Function[varAssignments,
						(* replace Boolean variables with their fixed values;
							the basic logic operations (And/Or/...) will simplify themselves,
							but Equal requires manual handling:
						*)
						tempExpr=noDomainsExpr/.varAssignments;
						tempExpr=tempExpr/.{
							Equal[True,st_]->st,
							Equal[st_,True]->st,
							Equal[False,st_]->Not[st],
							Equal[st_,False]->Not[st]
						};
						
						tempResults=findInstanceUniqueVars[
							tempExpr,
							varInts,
							Integers
						];
						
						If[tempResults!={},
							AppendTo[results,
								Join[First[tempResults],varAssignments]
							];
							
							(* one solution is enough here: *)
							Throw[False]; 
						]
					],
					boolTuples
				]
			]
		]
	];
	
	(* remove temp variables that were used for IntegerPart[x] replacements;
		the pattern for deletion is tempVar1|tempVar2|tempVar3\[Rule]_ *)
	results=DeleteCases[results,Alternatives@@ipVars->_,All];
	Scan[Remove,ipVars];
	
	results
]


End[];


EndPackage[];
